#!/usr/bin/env python3

import collections
import contextlib
import math
import os
import re
import shutil
import threading
import time
import warnings
from bisect import bisect_left, bisect_right
from collections import defaultdict
from contextlib import contextmanager
from functools import partial, wraps
from glob import glob

import mngs
import numpy as np
import pandas as pd
import readchar
import torch
from natsort import natsorted


################################################################################
## strings
################################################################################
def decapitalize(s):
    if not s:  # check that s is not empty string
        return s
    return s[0].lower() + s[1:]


def connect_strs(strs, filler="_"):  # connect_nums also works as connect_strs
    """
    Example:
        print(connect_strs(['a', 'b', 'c'], filler='_'))
        # 'a_b_c'
    """
    if isinstance(strs, list) or isinstance(strs, tuple):
        connected = ""
        for s in strs:
            connected += filler + s
        return connected[len(filler) :]
    else:
        return strs


def connect_nums(nums, filler="_"):
    """
    Example:
        print(connect_nums([1, 2, 3], filler='_'))
        # '1_2_3'
    """
    if isinstance(nums, list) or isinstance(nums, tuple):
        connected = ""
        for n in nums:
            connected += filler + str(n)
        return connected[len(filler) :]
    else:
        return nums


def squeeze_spaces(string, pattern=" +", repl=" "):
    """Return the string obtained by replacing the leftmost
    non-overlapping occurrences of the pattern in string by the
    replacement repl.  repl can be either a string or a callable;
    if a string, backslash escapes in it are processed.  If it is
    a callable, it's passed the Match object and must return
    a replacement string to be used.
    """
    # return re.sub(" +", " ", string)
    return re.sub(pattern, repl, string)


import re

import numpy as np


def search(patterns, strings, only_perfect_match=False, as_bool=False):
    """
    regular expression is acceptable for patterns.

    Example:
        patterns = ['orange', 'banana']
        strings = ['apple', 'orange', 'apple', 'apple_juice', 'banana', 'orange_juice']
        print(search(patterns, strings))
        # ([1, 4, 5], ['orange', 'banana', 'orange_juice'])

        patterns = 'orange'
        strings = ['apple', 'orange', 'apple', 'apple_juice', 'banana', 'orange_juice']
        print(search(patterns, strings))
        # ([1, 5], ['orange', 'orange_juice'])
    """

    ## For single string objects
    def to_list(s_or_p):
        if isinstance(s_or_p, collections.abc.KeysView):
            s_or_p = list(s_or_p)

        elif not isinstance(
            s_or_p,
            (list, tuple, pd.core.indexes.base.Index, pd.core.series.Series),
        ):
            s_or_p = [s_or_p]

        return s_or_p

    patterns = to_list(patterns)
    strings = to_list(strings)

    ## Main
    if not only_perfect_match:
        indi_matched = []
        for pattern in patterns:
            for i_str, string in enumerate(strings):
                m = re.search(pattern, string)
                if m is not None:
                    indi_matched.append(i_str)
    else:
        indi_matched = []
        for pattern in patterns:
            for i_str, string in enumerate(strings):
                if pattern == string:
                    indi_matched.append(i_str)

    ## Sorts the indices according to the original strings
    indi_matched = natsorted(indi_matched)
    keys_matched = list(np.array(strings)[indi_matched])

    if as_bool:
        bool_matched = np.zeros(len(strings), dtype=bool)
        if np.unique(indi_matched).size != 0:
            bool_matched[np.unique(indi_matched)] = True
        return bool_matched, keys_matched

    else:
        return indi_matched, keys_matched


# def search(patterns, strings, only_perfect_match=False, as_bool=False):
#     """
#     regular expression is acceptable for patterns.

#     Example:
#         patterns = ['orange', 'banana']
#         strings = ['apple', 'orange', 'apple', 'apple_juice', 'banana', 'orange_juice']
#         print(search(patterns, strings))
#         # ([1, 4, 5], ['orange', 'banana', 'orange_juice'])

#         patterns = 'orange'
#         strings = ['apple', 'orange', 'apple', 'apple_juice', 'banana', 'orange_juice']
#         print(search(patterns, strings))
#         # ([1, 5], ['orange', 'orange_juice'])
#     """

#     ## For single string objects
#     def to_str_list(data):
#         data_arr = np.array(data).astype(str)
#         if data_arr.ndim == 0:
#             data_arr = data_arr[np.newaxis]
#         return list(data_arr)

#     # def to_list(s_or_p):
#     #     if isinstance(s_or_p, collections.abc.KeysView):
#     #         s_or_p = list(s_or_p)

#     #     elif not isinstance(
#     #         s_or_p,
#     #         (list, tuple, pd.core.indexes.base.Index, pd.core.series.Series),
#     #     ):
#     #         s_or_p = [s_or_p]

#     #     return s_or_p

#     # patterns = to_list(patterns)
#     # strings = to_list(strings)
#     patterns = to_str_list(patterns)
#     strings = to_str_list(strings)

#     ## Main
#     if not only_perfect_match:
#         indi_matched = []
#         for pattern in patterns:
#             for i_str, string in enumerate(strings):
#                 m = re.search(pattern, string)
#                 if m is not None:
#                     indi_matched.append(i_str)
#     else:
#         indi_matched = []
#         for pattern in patterns:
#             for i_str, string in enumerate(strings):
#                 if pattern == string:
#                     indi_matched.append(i_str)

#     ## Sorts the indices according to the original strings
#     indi_matched = natsorted(indi_matched)
#     keys_matched = list(np.array(strings)[indi_matched])

#     if as_bool:
#         bool_matched = np.zeros(len(strings), dtype=bool)
#         if np.unique(indi_matched).size != 0:
#             bool_matched[np.unique(indi_matched)] = True
#         return bool_matched, keys_matched

#     else:
#         return indi_matched, keys_matched


def grep(str_list, search_key):
    """
    Example:
        str_list = ['apple', 'orange', 'apple', 'apple_juice', 'banana', 'orange_juice']
        search_key = 'orange'
        print(grep(str_list, search_key))
        # ([1, 5], ['orange', 'orange_juice'])
    """
    matched_keys = []
    indi = []
    for ii, string in enumerate(str_list):
        m = re.search(search_key, string)
        if m is not None:
            matched_keys.append(string)
            indi.append(ii)
    return indi, matched_keys


def pop_keys(keys_list, keys_to_pop):
    """
    keys_list = ['a', 'b', 'c', 'd', 'e', 'bde']
    keys_to_pop = ['b', 'd']
    pop_keys(keys_list, keys_to_pop)
    """
    indi_to_remain = [k not in keys_to_pop for k in keys_list]
    keys_remainded_list = list(np.array(keys_list)[list(indi_to_remain)])
    return keys_remainded_list


def readable_bytes(num, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return "%3.1f %s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f %s%s" % (num, "Yi", suffix)


################################################################################
## list
################################################################################
def is_listed_X(obj, types):
    """
    Example:
        obj = [3, 2, 1, 5]
        _is_listed_X(obj,
    """
    import numpy as np

    try:
        conditions = []
        condition_list = isinstance(obj, list)

        if not (isinstance(types, list) or isinstance(types, tuple)):
            types = [types]

        _conditions_susp = []
        for typ in types:
            _conditions_susp.append(
                (np.array([isinstance(o, typ) for o in obj]) == True).all()
            )

        condition_susp = np.any(_conditions_susp)

        _is_listed_X = np.all([condition_list, condition_susp])
        return _is_listed_X

    except:
        return False


def find_closest(list_obj, num_insert):
    """
    Assumes list_obj is sorted. Returns the closest value to num.
    If the same number is included in list_obj, the smaller number is returned.

    Example:
        list_obj = np.array([0, 1, 1, 2, 3, 3])
        num = 1.2
        closest_num, closest_pos = take_the_closest(list_obj, num)
        print(closest_num, closest_pos)
        # 1 2

        list_obj = np.array([0, 1, 1, 2, 3, 3])
        num = 1
        closest_num, closest_pos = take_the_closest(list_obj, num)
        print(closest_num, closest_pos)
        # 1 1
    """
    if math.isnan(num_insert):
        closest_num = np.nan
        closest_pos = np.nan

    pos_num_insert = bisect_left(list_obj, num_insert)

    if pos_num_insert == 0:  # When the insertion is at the first position
        closest_num = list_obj[0]
        closest_pos = pos_num_insert  # 0

    if pos_num_insert == len(
        list_obj
    ):  # When the insertion is at the last position
        closest_num = list_obj[-1]
        closest_pos = pos_num_insert  # len(list_obj)

    else:  # When the insertion is anywhere between the first and the last positions
        pos_before = pos_num_insert - 1

        before_num = list_obj[pos_before]
        after_num = list_obj[pos_num_insert]

        delta_after = abs(after_num - num_insert)
        delta_before = abs(before_num - num_insert)

        if np.abs(delta_after) < np.abs(delta_before):
            closest_num = after_num
            closest_pos = pos_num_insert

        else:
            closest_num = before_num
            closest_pos = pos_before

    return closest_num, closest_pos


################################################################################
## mutable
################################################################################
def isclose(mutable_a, mutable_b):
    return [math.isclose(a, b) for a, b in zip(mutable_a, mutable_b)]


################################################################################
## dictionary
################################################################################
def merge_dicts_wo_overlaps(*dicts):
    merged_dict = {}  # init
    for dict in dicts:
        assert mngs.general.search(
            merged_dict.keys(), dict.keys(), only_perfect_match=True
        ) == ([], [])
        merged_dict.update(dict)
    return merged_dict


def listed_dict(keys=None):  # Is there a better name?
    """
    Example 1:
        import random
        random.seed(42)
        d = listed_dict()
        for _ in range(10):
            d['a'].append(random.randint(0, 10))
        print(d)
        # defaultdict(<class 'list'>, {'a': [10, 1, 0, 4, 3, 3, 2, 1, 10, 8]})

    Example 2:
        import random
        random.seed(42)
        keys = ['a', 'b', 'c']
        d = listed_dict(keys)
        for _ in range(10):
            d['a'].append(random.randint(0, 10))
            d['b'].append(random.randint(0, 10))
            d['c'].append(random.randint(0, 10))
        print(d)
        # defaultdict(<class 'list'>, {'a': [10, 4, 2, 8, 6, 1, 8, 8, 8, 7],
        #                              'b': [1, 3, 1, 1, 0, 3, 9, 3, 6, 9],
        #                              'c': [0, 3, 10, 9, 0, 3, 0, 10, 3, 4]})
    """
    dict_list = defaultdict(list)
    # initialize with keys if possible
    if keys is not None:
        for k in keys:
            dict_list[k] = []
    return dict_list


################################################################################
## variables
################################################################################
def is_defined_global(x_str):
    """
    Example:
        print(is_defined('a'))
        # False

        a = 5
        print(is_defined('a'))
        # True
    """
    return x_str in globals()


def is_defined_local(x_str):
    """
    Example:
        print(is_defined('a'))
        # False

        a = 5
        print(is_defined('a'))
        # True
    """
    return x_str in locals()


# def does_exist(suspicious_var_str):
#     return suspicious_var_str in globals()


################################################################################
## versioning
################################################################################
def is_later_or_equal(package, tgt_version, format="MAJOR.MINOR.PATCH"):
    import mngs
    import numpy as np

    indi, matched = mngs.general.search(
        ["MAJOR", "MINOR", "PATCH"], format.split(".")
    )
    imp_major, imp_minor, imp_patch = [
        int(v) for v in np.array(package.__version__.split("."))[indi]
    ]
    tgt_major, tgt_minor, tgt_patch = [
        int(v) for v in np.array(tgt_version.split("."))[indi]
    ]

    print(
        f"\npackage: {package.__name__}\n"
        f"target_version: {tgt_version}\n"
        f"imported_version: {imp_major}.{imp_minor}.{imp_patch}\n"
    )

    ## Mjorr
    if imp_major > tgt_major:
        return True

    if imp_major < tgt_major:
        return False

    if imp_major == tgt_major:

        ## Minor
        if imp_minor > tgt_minor:
            return True

        if imp_minor < tgt_minor:
            return False

        if imp_minor == tgt_minor:

            ## Patch
            if imp_patch > tgt_patch:
                return True
            if imp_patch < tgt_patch:
                return False
            if imp_patch == tgt_patch:
                return True


################################################################################
## File
################################################################################
def _copy_a_file(src, dst, allow_overwrite=False):
    if src == "/dev/null":
        print(f"\n/dev/null was not copied.\n")

    else:

        if dst.endswith("/"):
            _, src_fname, src_ext = mngs.path.split(src)
            # src_fname = src + src_ext
            dst = dst + src_fname + src_ext

        if not os.path.exists(dst):
            shutil.copyfile(src, dst)
            print(f'\nCopied "{src}" to "{dst}".\n')

        else:
            if allow_overwrite:
                shutil.copyfile(src, dst)
                print(f'\nCopied "{src}" to "{dst}" (overwritten).\n')

            if not allow_overwrite:
                print(
                    f'\n"{dst}" exists and copying from "{src}" was aborted.\n'
                )


def copy_files(src_files, dists, allow_overwrite=False):
    if isinstance(src_files, str):
        src_files = [src_files]

    if isinstance(dists, str):
        dists = [dists]

    for sf in src_files:
        for dst in dists:
            _copy_a_file(sf, dst, allow_overwrite=allow_overwrite)


def copy_the_file(sdir):  # dst
    __file__ = inspect.stack()[1].filename
    _, fname, ext = mngs.path.split(__file__)

    dst = sdir + fname + ext

    if "ipython" not in __file__:  # for ipython
        # shutil.copyfile(__file__, dst)
        # print(f"Saved to: {dst}")
        _copy_a_file(__file__, dst)


def is_nan(X):
    if isinstance(X, pd.DataFrame):
        if X.isna().any().any():
            raise ValueError("NaN was found in X")
    elif isinstance(X, np.ndarray):
        if np.isnan(X).any():
            raise ValueError("NaN was found in X")
    elif torch.is_tensor(X):
        if X.isnan().any():
            raise ValueError("NaN was found in X")
    elif isinstance(X, (float, int)):
        if math.isnan(X):
            raise ValueError("X was NaN")


def partial_at(func, index, value):
    @wraps(func)
    def result(*rest, **kwargs):
        args = []
        args.extend(rest[:index])
        args.append(value)
        args.extend(rest[index:])
        return func(*args, **kwargs)

    return result


def describe(df, method="mean", factor=1):
    df = pd.DataFrame(df)
    # df = df[~df[0].isna()]
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        if method == "mean":
            return round(np.nanmean(df), 3), round(np.nanstd(df) / factor, 3)
        if method == "median":
            med = df.describe().T["50%"].iloc[0]
            IQR = (
                df.describe().T["75%"].iloc[0] - df.describe().T["25%"].iloc[0]
            )
            return round(med, 3), round(IQR / factor, 3)


# def describe(arr, method="mean", factor=1):
#     arr = pd.DataFrame(arr)
#     arr = np.hstack(arr[~np.isnan(arr)])

#     with warnings.catch_warnings():
#         warnings.simplefilter("ignore", RuntimeWarning)
#         if method == "mean":
#             return np.nanmean(df), np.nanstd(df) / factor
#         if method == "median":
#             med = df.describe().T["50%"].iloc[0]
#             IQR = df.describe().T["75%"].iloc[0] - df.describe().T["25%"].iloc[0]
#             return med, IQR / factor

# def count():
#     counter = 0
#     while True:
#         print(counter)
#         time.sleep(1)
#         counter += 1
def _return_counting_process():
    import multiprocessing

    def _count():
        counter = 0
        while True:
            print(counter)
            time.sleep(1)
            counter += 1

    p1 = multiprocessing.Process(target=_count)
    p1.start()
    return p1


def wait_key(process, tgt_key="q"):
    """
    Example:
        import mngs
        p1 = mngs.general._return_counting_process()
        mngs.gen.wait_key(p1)
        # press q
    """
    pressed_key = None
    while pressed_key != tgt_key:  # "q"
        pressed_key = readchar.readchar()
        print(pressed_key)
    process.terminate()


class ThreadWithReturnValue(threading.Thread):
    """
    Example:
        t = ThreadWithReturnValue(
            target=func, args=(,), kwargs={key: val}
        )
        t.start()
        out = t.join()

    """

    def __init__(
        self,
        group=None,
        target=None,
        name=None,
        args=(),
        kwargs={},
        Verbose=None,
    ):
        Thread.__init__(self, group, target, name, args, kwargs)
        self._return = None

    def run(self):
        if self._target is not None:
            self._return = self._target(*self._args, **self._kwargs)

    def join(self, *args):
        ### fixme
        Thread.join(self, *args)
        return self._return


# @contextmanager
# def suppress_output():
#     """A context manager that suppresses stdout and stderr."""
#     with open(os.devnull, "w") as fnull:
#         with contextlib.redirect_stdout(fnull), contextlib.redirect_stderr(
#             fnull
#         ):
#             yield
@contextmanager
def suppress_output():
    """
    A context manager that suppresses stdout and stderr.

    Example:
        with suppress_output():
            print("This will not be printed to the console.")
    """
    # Open a file descriptor that points to os.devnull (a black hole for data)
    with open(os.devnull, "w") as fnull:
        # Temporarily redirect stdout to the file descriptor fnull
        with contextlib.redirect_stdout(fnull):
            # Temporarily redirect stderr to the file descriptor fnull
            with contextlib.redirect_stderr(fnull):
                # Yield control back to the context block
                yield


def unique(data, axis=None):
    """
    Identifies unique elements in the data along the specified axis and their counts, returning a DataFrame.

    Parameters:
    - data (array-like): The input data to analyze for unique elements.
    - axis (int, optional): The axis along which to find the unique elements. Defaults to None.

    Returns:
    - df (pandas.DataFrame): DataFrame with unique elements and their counts.
    """
    if axis is None:
        uqs, counts = np.unique(data, return_counts=True)
    else:
        uqs, counts = np.unique(data, axis=axis, return_counts=True)

    if axis is None:
        df = pd.DataFrame({"uq": uqs, "n": counts})
    else:
        df = pd.DataFrame(
            uqs, columns=[f"axis_{i}" for i in range(uqs.shape[1])]
        )
        df["n"] = counts

    df["n"] = df["n"].apply(lambda x: f"{int(x):,}")

    return df


def unique(data, axis=None):
    """
    Identifies unique elements in the data along the specified axis and their counts, returning a DataFrame.

    Parameters:
    - data (array-like): The input data to analyze for unique elements.
    - axis (int, optional): The axis along which to find the unique elements. Defaults to None.

    Returns:
    - df (pandas.DataFrame): DataFrame with unique elements and their counts.
    """
    # Find unique elements and their counts
    if axis is None:
        uqs, counts = np.unique(data, return_counts=True)
        df = pd.DataFrame({"Unique Elements": uqs, "Counts": counts})
    else:
        uqs, counts = np.unique(data, axis=axis, return_counts=True)
        # Create a DataFrame with unique elements
        df = pd.DataFrame(
            uqs,
            columns=[f"Unique Elements Axis {i}" for i in range(uqs.shape[1])],
        )
        # Add a column for counts
        df["Counts"] = counts

    # Format the 'Counts' column with commas for thousands
    df["Counts"] = df["Counts"].apply(lambda x: f"{x:,}")

    return df


def uq(*args, **kwargs):
    return unique(*args, **kwargs)


# def uq(data, axis=None):
#     def _uq(data):
#         uqs, counts = np.unique(data, return_counts=True)
#         df = pd.DataFrame({"uq": uqs, "n": counts})
#         # Format the 'Counts' column with commas for thousands
#         df["n"] = df["n"].apply(lambda x: f"{x:,}")
#         return df

#     data = pd.DataFrame(data)

#     if axis == 1:
#         dfs = {}
#         for col in data.columns:
#             df = _uq(data[col])
#             dfs[col] = df
#         return dfs

#     if axis == 0:
#         dfs = {}
#         for col in data.T.columns:
#             df = _uq(data.T[col])
#             dfs[col] = df
#         return dfs

#     if axis is None:
#         return _uq(data)


# def unique(data, axis=None):
#     """
#     Identifies unique elements in the data and their counts, returning a DataFrame.

#     Parameters:
#     - data (array-like): The input data to analyze for unique elements.
#     - show (bool, optional): If True, prints the DataFrame. Defaults to True.

#     Returns:
#     - df (pandas.DataFrame): DataFrame with unique elements and their counts.
#     """
#     uqs, counts = np.unique(data, return_counts=True)  # [REVISED]
#     df = pd.DataFrame(
#         np.vstack([uqs, counts]).T, columns=["uq", "n"]  # [REVISED]
#     ).set_index(
#         "uq"
#     )  # [REVISED]

#     df_show = df.copy()
#     df_show["n"] = df_show["n"].apply(lambda x: f"{int(x):,}")  # [REVISED]

#     return df_show
def print_block(message, char="-", n=40, c=None):
    """Available colors are 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white', or 'grey'"""
    border = char * n
    text = f"\n{border}\n{message}\n{border}\n"
    if c is not None:
        text = color_text(text, c)
    print(text)


def color_text(text, c="green"):
    ANSI_COLORS = {
        "red": "\033[91m",
        "green": "\033[92m",
        "yellow": "\033[93m",
        "blue": "\033[94m",
        "magenta": "\033[95m",
        "cyan": "\033[96m",
        "white": "\033[97m",
        "grey": "\033[90m",  # Added grey color
        "gray": "\033[90m",  # Added grey color
        "reset": "\033[0m",
    }
    ANSI_COLORS["tra"] = ANSI_COLORS["white"]
    ANSI_COLORS["val"] = ANSI_COLORS["green"]
    ANSI_COLORS["tes"] = ANSI_COLORS["red"]

    start_code = ANSI_COLORS.get(c, ANSI_COLORS["reset"])
    end_code = ANSI_COLORS["reset"]
    return f"{start_code}{text}{end_code}"


ct = color_text


def mv_col(dataframe, column_name, position):
    temp_col = dataframe[column_name]
    dataframe.drop(labels=[column_name], axis=1, inplace=True)
    dataframe.insert(loc=position, column=column_name, value=temp_col)
    return dataframe


def symlink(tgt, src, force=False):
    if force:
        try:
            os.remove(src)
        except FileNotFoundError:
            pass

    # Calculate the relative path from src to tgt
    src_dir = os.path.dirname(src)
    relative_tgt = os.path.relpath(tgt, src_dir)

    os.symlink(relative_tgt, src)
    print(
        mngs.gen.ct(
            f"\nSymlink was created: {src} -> {relative_tgt}\n", c="yellow"
        )
    )


#     os.symlink(tgt, src)
#     print(mngs.gen.ct(f"\nSymlink was created: {src} -> {tgt}\n", c="yellow"))
# Symlink was created: ./scripts/ml/clf/sct_optuna/optuna_studies/optuna_study_stent_3_classes/best_trial -> /home/ywatanabe/proj/ecog_stent_sheep_visual/scripts/ml/clf/sct_optuna/RUNNING/2024Y-03M-29D-21h55m09s_IBSy/objective/Trial#00068/


def to_even(n):
    if n % 2 == 0:
        return n
    else:
        return n - 1


def to_odd(n):
    if n % 2 == 0:
        return n - 1
    else:
        return n


def natglob(expression):
    return natsorted(glob(expression))



import numpy as np
from scipy import stats


def smirnov_grubbs(data_arr, alpha=0.05):
    """
    Find outliers based on Smirnov-Grubbs test.

    Arguments:

    Returns | indices of outliers
    """
    data_1D_sorted = sorted(np.array(data_arr).reshape(-1))
    in_data, out_data = list(data_1D_sorted), []

    # while True:
    n = len(in_data)
    for i_data in range(n):
        n = len(in_data)
        t = stats.t.isf(q=(alpha / n) / 2, df=n - 2)
        tau = (n - 1) * t / np.sqrt(n * (n - 2) + n * t * t)
        i_min, i_max = np.argmin(in_data), np.argmax(in_data)
        mu, std = np.mean(in_data), np.std(in_data, ddof=1)

        i_far = (
            i_max
            if np.abs(in_data[i_max] - mu) > np.abs(in_data[i_min] - mu)
            else i_min
        )

        tau_far = np.abs((in_data[i_far] - mu) / std)

        if tau_far < tau:
            break

        out_data.append(in_data.pop(i_far))

    if len(out_data) == 0:
        return None

    else:
        out_data_uq = np.unique(out_data)
        indi_outliers = np.hstack(
            [
                np.vstack(np.where(data_arr == out_data_uq[i_out])).T
                for i_out in range(len(out_data_uq))
            ]
        ).squeeze()

        if indi_outliers.ndim == 0:
            indi_outliers = indi_outliers[np.newaxis]
        return indi_outliers


# def smirnov_grubbs(data_arr, alpha=0.05):
#     """
#     Find outliers based on Smirnov-Grubbs test.

#     Arguments:

#     Returns | indices of outliers
#     """
#     data_1D_sorted = sorted(np.array(data_arr).reshape(-1))
#     in_data, out_data = list(data_1D_sorted), []

#     while True:
#         n = len(in_data)
#         t = stats.t.isf(q=(alpha / n) / 2, df=n - 2)
#         tau = (n - 1) * t / np.sqrt(n * (n - 2) + n * t * t)
#         i_min, i_max = np.argmin(in_data), np.argmax(in_data)
#         mu, std = np.mean(in_data), np.std(in_data, ddof=1)

#         i_far = (
#             i_max
#             if np.abs(in_data[i_max] - mu) > np.abs(in_data[i_min] - mu)
#             else i_min
#         )

#         tau_far = np.abs((in_data[i_far] - mu) / std)

#         if tau_far < tau:
#             break

#         out_data.append(in_data.pop(i_far))

#     if len(out_data) == 0:
#         return None

#     else:
#         out_data_uq = np.unique(out_data)
#         indi_outliers = np.vstack(
#             [
#                 np.vstack(np.where(data_arr == out_data_uq[i_out])).T
#                 for i_out in range(len(out_data_uq))
#             ]
#         )

#         return np.array(indi_outliers).squeeze()



"""
This script does XYZ.
"""

import os
import sys

import matplotlib.pyplot as plt
import mngs

# Imports
import numpy as np
import pandas as pd
import sounddevice as sd
import torch
import torch.nn as nn
import torch.nn.functional as F

# Config
CONFIG = mngs.gen.load_configs()

# Functions
def listen(x, fs, chs=(0, 1)):
    """
    Play selected channels of a multichannel signal array as audio.

    Parameters:
    - x (numpy.ndarray): Signal array of shape (batch_size, n_chs, seq_len).
    - fs (int): Sampling frequency of the signal.
    - chs (tuple of int): Tuple of channel indices to listen to.

    Returns:
    - None

    Memo:
        The PortAudio package is required.
        sudo yum update && sudo yum install -y portaudio portaudio-devel
    """
    COMMON_FS = 44100

    # Ensure the input channels are within the range of available channels
    if max(chs) >= x.shape[1]:
        raise ValueError("Channel index out of range")

    # Extract the desired channels (average if more than one)
    selected_channels = x[:, chs, :].mean(axis=1)

    # Flatten the batch dimension by averaging (or any other method as necessary)
    signal_to_play = selected_channels.mean(axis=0)

    # Resample the signal to the common sample rate if necessary
    if fs != COMMON_FS:
        from scipy.signal import resample

        num_samples = int(round(len(signal_to_play) * common_fs / fs))
        signal_to_play = resample(signal_to_play, num_samples)

    # Play the sound
    sd.play(signal_to_play, fs)
    sd.wait()  # Wait until the sound is finished playing


def print_device_info():
    device_info = sd.query_devices(kind="output")
    print(f"Default Output Device Info: \n{device_info}")


def list_and_select_device():
    print("Available audio devices:")
    print(sd.query_devices())
    device_id = int(input("Enter the ID of the device you want to use: "))
    return device_id


if __name__ == "__main__":
    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(sys, plt)

    xx, tt, fs = mngs.dsp.demo_sig("chirp")

    device_id = list_and_select_device()
    sd.default.device = device_id

    listen(xx, common_fs)

    # Close
    mngs.gen.close(CONFIG)

# EOF

"""
/ssh:ywatanabe@444:/home/ywatanabe/proj/entrance/mngs/dsp/_listen.py
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-02-26 16:32:42 (ywatanabe)"

import warnings

import numpy as np
import torch
from sklearn.metrics import balanced_accuracy_score


def bACC(true_class, pred_class):
    """
    Calculates the balanced accuracy score between predicted and true class labels.

    Parameters:
    - true_class (array-like or torch.Tensor): True class labels.
    - pred_class (array-like or torch.Tensor): Predicted class labels.

    Returns:
    - bACC (float): The balanced accuracy score rounded to three decimal places.
    """
    if isinstance(true_class, torch.Tensor):  # [REVISED]
        true_class = true_class.detach().cpu().numpy()  # [REVISED]
    if isinstance(pred_class, torch.Tensor):  # [REVISED]
        pred_class = pred_class.detach().cpu().numpy()  # [REVISED]

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        bACC_score = balanced_accuracy_score(
            true_class.reshape(-1),  # [REVISED]
            pred_class.reshape(-1),  # [REVISED]
        )
    return round(bACC_score, 3)  # [REVISED]



#!/usr/bin/env python3


def to_the_latex_style(str_or_num):
    """
    Example:
        print(to_the_latex_style('aaa'))
        # '$aaa$'
    """
    string = str(str_or_num)
    if (string[0] == "$") and (string[-1] == "$"):
        return string
    else:
        return "${}$".format(string)


def add_hat_in_the_latex_style(str_or_num):
    """
    Example:
        print(add_hat_in_the_latex_style('aaa'))
        # '$\\hat{aaa}$'
    """
    return to_the_latex_style("\hat{%s}" % str_or_num)



#!/usr/bin/env python3

import numpy as np
from scipy import stats


def calc_partial_corrcoef(x, y, z):
    """remove the influence of the variable z from the correlation between x and y."""
    r_xy = np.corrcoef(x, y)
    r_xz = np.corrcoef(x, z)
    r_yz = np.corrcoef(y, z)
    r_xy_z = (r_xy - r_xz * r_yz) / (1 - r_xz ** 2) * (1 - r_yz ** 2)
    return r_xy_z


def nocorrelation_test(x, y, z=None, alpha=0.05):
    if z is None:
        r = np.corrcoef(x, y)[1, 0]
    if z is not None:
        r = calc_partial_corrcoef(x, y, z)[1, 0]

    n = len(x)
    df = n - 2
    # t = np.abs(np.array(r)) * np.sqrt((df) / (1 - np.array(r)**2))
    t = np.abs(r) * np.sqrt((df) / (1 - r ** 2))
    # t_alpha = scipy.stats.t.ppf(1 - alpha / 2, df)
    p_value = 2 * (1 - stats.t.cdf(t, df))
    return r, t, p_value



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-15 02:11:56 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/ml/chat.py


"""
This script does XYZ.
"""


"""
Imports
"""
# mngs.gen.reload(mngs)
import io
import json
import os
import re
import sys
import warnings
from abc import ABC, abstractmethod
from glob import glob
from pprint import pprint

import anthropic
import google.generativeai as genai
import markdown2
import matplotlib
import matplotlib.pyplot as plt
import mngs
import numpy as np
import openai
import pandas as pd
import seaborn as sns
import torch
import torch.nn as nn
import torch.nn.functional as F
import xarray as xr
from icecream import ic

# from mngs.gen import notify
from natsort import natsorted
from openai import OpenAI
from tqdm import tqdm

from mngs.ai._gen_ai._ChatGPT import ChatGPT
from mngs.ai._gen_ai._Claude import Claude
from mngs.ai._gen_ai._Gemini import Gemini
from mngs.ai._gen_ai._Perplexity import Perplexity

# from sciwriter_app._email import notify

# sys.path = ["."] + sys.path
# from scripts import utils, load

"""
Warnings
"""
# warnings.simplefilter("ignore", UserWarning)


"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""


MODEL_CONFIG = {
    "ChatGPT": {
        "models": ["gpt-4o", "gpt-4-turbo", "gpt-4", "gpt-3.5-turbo"],
        "api_key_env": "OPENAI_API_KEY",
        "class": "ChatGPT",
    },
    "Claude": {
        "models": [
            "claude-3-opus-20240229",
            "claude-3-sonnet-20240229",
            "claude-3-haiku-20240307",
        ],
        "api_key_env": "CLAUDE_API_KEY",
        "class": "Claude",
    },
    "Gemini": {
        "models": [
            "gemini-1.5-pro-latest",
            "gemini-1.5-flash-latest",
            "gemini-1.5-pro",
            "gemini-pro",
        ],
        "api_key_env": "GOOGLE_API_KEY",
        "class": "Gemini",
    },
    "Perplexity": {
        "models": [
            "llama-3-sonar-small-32k-chat",
            "llama-3-sonar-small-32k-online",
            "llama-3-sonar-large-32k-chat",
            "llama-3-sonar-large-32k-online",
            "llama-3-8b-instruct",
            "llama-3-70b-instruct",
            "mixtral-8x7b-instruct",
        ],
        "api_key_env": "PERPLEXITY_API_KEY",
        "class": "Perplexity",
    },
}


def GenAI(
    model="gpt-3.5-turbo",
    stream=False,
    api_key=None,
    seed=None,
    temperature=1.0,
    n_keep=1,
):
    """Factory function to create an instance of an AI model handler."""
    if api_key is None:
        api_key = model2apikey(model)

    for config in MODEL_CONFIG.values():
        if model in config["models"]:
            model_class = globals()[config["class"]]
            return model_class(
                model=model,
                stream=stream,
                api_key=api_key,
                seed=seed,
                temperature=temperature,
                n_keep=n_keep,
            )

    raise ValueError(f"No handler available for model {model}.")


################################################################################
# Helper functions
################################################################################
def model2apikey(model):
    """Retrieve the API key for a given model from environment variables."""
    for config in MODEL_CONFIG.values():
        if model in config["models"]:
            api_key = os.getenv(config["api_key_env"])
            if not api_key:
                raise EnvironmentError(
                    f"API key for {model} not found in environment."
                )
            return api_key
    raise ValueError(f"Model {model} is not supported.")


def test_all(seed=None, temperature=1.0):
    model_names = [
        "gpt-4",
        "claude-3-opus-20240229",
        "gemini-pro",
        "llama-3-sonar-large-32k-online",
    ]

    for model_name in model_names:
        for stream in [False, True]:
            model = GenAI(
                model_name, stream=stream, seed=seed, temperature=temperature
            )
            # prompt = "Hi. Tell me your name just within a line."
            prompt = "Hi. Tell me about the hippocampus."            

            print(
                f"\n{'-'*40}\n{model.model}\nStream: {stream}\nSeed: {seed}\nTemperature: {temperature}\n{'-'*40}"
            )
            print(model(prompt))
            print(model.available_models)


def main(
    model="gpt-3.5-turbo",
    stream=False,
    prompt="Hi, please tell me about the hippocampus",
    seed=None,
    temperature=1.0,
):

    m = GenAI(model, stream=stream, seed=seed, temperature=temperature)
    out = m(prompt)

    return out


if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    test_all()
    # out_static = main(stream=False, seed=42)
    # out_static = main(stream=False, seed=42)

    # out_stream = main(stream=True)

    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-06 08:36:13 (ywatanabe)"

import inspect
import os as _os
from datetime import datetime
from glob import glob
from pprint import pprint
from time import sleep

import matplotlib
import mngs


def start(
    sys=None,
    plt=None,
    sdir=None,
    verbose=True,
    # Random seeds
    os=None,
    random=None,
    np=None,
    torch=None,
    tf=None,
    seed=42,
    # matplotlib
    agg=False,
    fig_size_mm=(160, 100),
    fig_scale=1.0,
    dpi_display=100,
    dpi_save=300,
    font_size_base=10,
    font_size_title=10,
    font_size_axis_label=10,
    font_size_tick_label=8,
    font_size_legend=8,
    hide_top_right_spines=True,
    alpha=0.9,
    line_width=0.5,
):
    """
    Example:

    \"""
    This script does XYZ.
    \"""

    # Imports
    import mngs
    import matplotlib.pyplot as plt
    import numpy as np
    import pandas as pd
    import torch
    import torch.nn as nn
    import torch.nn.functional as F

    # Config
    CONFIG = mngs.gen.load_configs()

    # Functions
    # (Your awesome code here)

    if __name__ == '__main__':
        # Start
        CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(sys, plt)

        # (Your awesome code here)

        # Close
        mngs.gen.close(CONFIG)

    # EOF

    \"""
    /home/ywatanabe/template.py
    \"""
    """

    # Timer
    start_time = datetime.now()

    # Debug mode check
    try:
        IS_DEBUG_PATH = "./config/IS_DEBUG.yaml"
        if _os.path.exists(IS_DEBUG_PATH):
            IS_DEBUG = mngs.io.load(IS_DEBUG_PATH).get("IS_DEBUG", False)
        else:
            IS_DEBUG = False

    except Exception as e:
        print(e)
        IS_DEBUG = False

    # ID
    ID = mngs.gen.gen_ID(N=4)
    ID = ID if not IS_DEBUG else "DEBUG_" + ID
    print(f"\n{'#'*40}\n## {ID}\n{'#'*40}\n")
    sleep(1)

    # Defines SDIR
    if sdir is None:
        __file__ = inspect.stack()[1].filename
        if "ipython" in __file__:
            __file__ = f"/tmp/fake_{_os.getenv('USER')}.py"
        spath = __file__
        _sdir, sfname, _ = mngs.path.split(spath)
        sdir = _sdir + sfname + "/" + "RUNNING" + "/" + ID + "/"
    _os.makedirs(sdir, exist_ok=True)

    # CONFIGs
    CONFIGS = mngs.io.load_configs(IS_DEBUG)
    CONFIGS["ID"] = ID
    CONFIGS["START_TIME"] = start_time
    CONFIGS["SDIR"] = sdir.replace("/./", "/")
    if verbose:
        print(f"\n{'-'*40}\n")
        print(f"CONFIG:")
        for k, v in CONFIGS.items():
            print(f"\n{k}:\n{v}\n")
        print(f"\n{'-'*40}\n")

    # Logging (tee)
    if sys is not None:
        sys.stdout, sys.stderr = mngs.general.tee(
            sys, sdir=sdir, verbose=verbose
        )

    # Random seeds
    if (
        (os is not None)
        or (random is not None)
        or (np is not None)
        or (torch is not None)
        or (tf is not None)
    ):
        mngs.gen.fix_seeds(
            os=os,
            random=random,
            np=np,
            torch=torch,
            tf=tf,
            seed=seed,
            verbose=verbose,
        )

    # Matplotlib configuration
    if plt is not None:
        plt, CC = mngs.plt.configure_mpl(
            plt,
            fig_size_mm=(160, 100),
            fig_scale=fig_scale,
            dpi_display=dpi_display,
            dpi_save=dpi_save,
            font_size_base=font_size_base,
            font_size_title=font_size_title,
            font_size_axis_label=font_size_axis_label,
            font_size_tick_label=font_size_tick_label,
            font_size_legend=font_size_legend,
            hide_top_right_spines=hide_top_right_spines,
            alpha=alpha,
            line_width=line_width,
            verbose=verbose,
        )
        CC["gray"] = CC["grey"]

    if agg:
        matplotlib.use("Agg")

    return CONFIGS, sys.stdout, sys.stderr, plt, CC


if __name__ == "__main__":
    """
    This script does XYZ.
    """

    # Imports
    import os
    import sys

    import matplotlib.pyplot as plt
    import mngs
    import numpy as np
    import pandas as pd
    import torch
    import torch.nn as nn
    import torch.nn.functional as F

    # Config
    CONFIG = mngs.gen.load_configs()

    # Functions
    # Your awesome code here :)

    if __name__ == "__main__":
        # Start
        CONFIG, sys.stdout, sys.stderr, plt, CC = start(sys, plt)

        # Your awesome code here :)

        # Close
        mngs.gen.close(CONFIG)

# EOF

"""
/home/ywatanabe/proj/entrance/mngs/general/_start.py
"""



#!/usr/bin/env python3

import numpy as np


def calc_partial_corr(x, y, z):
    """remove the influence of the variable z from the correlation between x and y."""

    x = np.array(x).astype(np.float128)
    y = np.array(y).astype(np.float128)
    z = np.array(z).astype(np.float128)

    r_xy = np.corrcoef(x, y)[0, 1]
    r_xz = np.corrcoef(x, z)[0, 1]
    r_yz = np.corrcoef(y, z)[0, 1]
    r_xy_z = (r_xy - r_xz * r_yz) / (np.sqrt(1 - r_xz ** 2) * np.sqrt(1 - r_yz ** 2))
    return r_xy_z



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-06 10:01:32 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/plt/_interp_colors.py


import matplotlib.colors as mcolors
import numpy as np


def interp_colors(color_start, color_end, num_points, round=3):
    # Convert colors to RGBA if they are not already in that format
    color_start_rgba = np.array(mcolors.to_rgba(color_start))
    color_end_rgba = np.array(mcolors.to_rgba(color_end))

    # Generate a sequence of RGBA values
    # np.linspace works on a per-component basis when the inputs are arrays
    rgba_values = np.linspace(
        color_start_rgba, color_end_rgba, num_points
    ).round(round)

    # Return the list of RGBA tuples
    return [list(color) for color in rgba_values]

    # # Generate a linear interpolation
    # return [for color in np.linspace(color_start_rgba, color_end_rgba, num_points)]
    # # return [
    # #     mcolors.to_hex(color)
    # #     for color in np.linspace(color_start_rgba, color_end_rgba, num_points)
    # # ]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-03-23 17:36:05 (ywatanabe)"

import numpy as np
from sklearn.decomposition import PCA
from sklearn.ensemble import GradientBoostingClassifier, RandomForestClassifier
from sklearn.feature_selection import SelectKBest, f_classif
from sklearn.linear_model import LogisticRegression, RidgeClassifierCV
from sklearn.pipeline import make_pipeline
from sklearn.svm import SVC, LinearSVC
from sktime.classification.deep_learning.cnn import CNNClassifier
from sktime.classification.deep_learning.inceptiontime import (
    InceptionTimeClassifier,
)
from sktime.classification.deep_learning.lstmfcn import LSTMFCNClassifier
from sktime.classification.dummy import DummyClassifier
from sktime.classification.feature_based import TSFreshClassifier
from sktime.classification.hybrid import HIVECOTEV2
from sktime.classification.interval_based import TimeSeriesForestClassifier
from sktime.classification.kernel_based import RocketClassifier, TimeSeriesSVC
from sktime.transformations.panel.reduce import Tabularizer
from sktime.transformations.panel.rocket import Rocket

# _rocket_pipeline = make_pipeline(
#     Rocket(n_jobs=-1),
#     RidgeClassifierCV(alphas=np.logspace(-3, 3, 10)),
# )


# def rocket_pipeline(*args, **kwargs):
#     return _rocket_pipeline


def rocket_pipeline(*args, **kwargs):
    return make_pipeline(
        Rocket(*args, **kwargs),
        LogisticRegression(
            max_iter=1000
        ),  # Increase max_iter if needed for convergence
        # RidgeClassifierCV(alphas=np.logspace(-3, 3, 10)),
        # SVC(probability=True, kernel="linear"),
    )


# def rocket_pipeline(*args, **kwargs):
#     return make_pipeline(
#         Rocket(*args, **kwargs),
#         SelectKBest(f_classif, k=500),
#         PCA(n_components=100),
#         LinearSVC(dual=False, tol=1e-3, C=0.1, probability=True),
#     )


GB_pipeline = make_pipeline(
    Tabularizer(),
    GradientBoostingClassifier(),
)



#!/usr/bin/env python3

import numpy as np

from ._PARAMS import PARAMS


def to_rgb(c):
    return PARAMS["RGB"][c]


def to_rgba(c, alpha=0.5):
    rgba = rgb2rgba(PARAMS["RGB"][c])
    rgba[-1] = alpha
    return rgba


def to_hex(c):
    return PARAMS["HEX"][c]


def rgb2rgba(rgb, alpha=0, round=2):
    rgb = np.array(rgb).astype(float)
    rgb /= 255
    return [*rgb.round(round), alpha]


def update_alpha(rgba, alpha):
    rgba_list = list(rgba)
    rgba_list[-1] = alpha
    return rgba_list


def rgba2hex(rgba):
    return "#{:02x}{:02x}{:02x}{:02x}".format(
        int(rgba[0]), int(rgba[1]), int(rgba[2]), int(rgba[3] * 255)
    )


def cycle_color(i_color):
    COLORS_10_STR = list(PARAMS["RGB"].keys())
    n_colors = len(COLORS_10_STR)
    return COLORS_10_STR[i_color % n_colors]


if __name__ == "__main__":
    c = "blue"
    print(to_rgb(c))
    print(to_rgba(c))
    print(to_hex(c))
    print(cycle_color(1))



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-03 11:23:10 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/dsp/_detect_ripples.py


"""
This script does XYZ.
"""


"""
Imports
"""
import sys

import matplotlib.pyplot as plt
import mngs

mngs.gen.reload(mngs)

import numpy as np
import pandas as pd
from scipy.signal import find_peaks

"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""


def _detect_ripples_preprocess(xx, fs, low_hz, high_hz, smoothing_sigma_ms=4):
    if xx.ndim == 2:
        xx = xx[np.newaxis]

    # Downsampling
    fs_tgt = low_hz * 3
    xx = mngs.dsp.resample(xx, float(fs), float(fs_tgt))
    fs = fs_tgt  # override

    # Subtracts the global mean to reduce false detection due to EMG signal
    xx -= np.nanmean(xx, axis=1, keepdims=True)

    RIPPLE_BANDS = np.vstack([[low_hz, high_hz]])

    # Bandpass Filtering
    xx = (
        (
            mngs.dsp.filt.bandpass(
                np.array(xx),
                fs,
                RIPPLE_BANDS,
            )
        )
        .squeeze(-2)
        .astype(np.float64)
    )

    # Calculate RMS
    xx = xx**2
    _, xx = mngs.dsp.hilbert(xx)
    xx = mngs.dsp.filt.gauss(xx, smoothing_sigma_ms * 1e-3 * fs).squeeze(-2)
    xx = np.sqrt(xx)

    # Scales across channels
    xx = xx.mean(axis=1)
    xx = mngs.gen.to_z(xx, dim=-1)
    return xx, fs_tgt


def detect_ripples(
    xx,
    fs,
    low_hz,
    high_hz,
    sd=2.0,
    smoothing_sigma_ms=4,
    min_duration_ms=10,
):
    """
    xx: 2-dimensional (n_chs, seq_len) or 3-dimensional (batch_size, n_chs, seq_len)
    """

    try:
        xx, _ = _detect_ripples_preprocess(
            xx, fs, low_hz, high_hz, smoothing_sigma_ms
        )

        # Detects peaks of RMS values
        dfs = []
        for ii in range(len(xx)):
            xi = xx[ii]

            # Finds peaks over the designated standard deviation
            peaks, properties = find_peaks(xi, height=sd)

            # Determines the range around each peak (customize as needed)
            peaks_all = []
            peak_ranges = []
            peak_amplitudes_sd = []

            for peak in peaks:
                left_bound = np.where(xi[:peak] < 0)[0]
                right_bound = np.where(xi[peak:] < 0)[0]

                left_ips = left_bound.max() if left_bound.size > 0 else peak
                right_ips = (
                    peak + right_bound.min() if right_bound.size > 0 else peak
                )

                # Avoid duplicates: Check if the current peak range is already listed
                if not any(
                    (left_ips == start and right_ips == end)
                    for start, end in peak_ranges
                ):
                    peaks_all.append(peak)
                    peak_ranges.append((left_ips, right_ips))
                    peak_amplitudes_sd.append(xi[peak])

            # Converts to DataFrame
            if peak_ranges:
                starts, ends = zip(*peak_ranges) if peak_ranges else ([], [])
                df = pd.DataFrame(
                    {
                        "start_s": np.hstack(starts) / fs,
                        "peak_s": np.hstack(peaks_all) / fs,
                        "end_s": np.hstack(ends) / fs,
                        "peak_amp_sd": np.hstack(peak_amplitudes_sd),
                    }
                ).round(3)
            else:
                df = pd.DataFrame(
                    columns=["start_s", "peak_s", "end_s", "peak_amp_sd"]
                )

            # Duration
            df["duration_s"] = df.end_s - df.start_s

            # Filters events with short duration
            df = df[df.duration_s > (min_duration_ms * 1e-3)]

            # Relative peak
            delta_s = df.peak_s - df.start_s
            rel_peak = delta_s / df.duration_s
            df["relative_peak_pos"] = np.round(rel_peak, 3)

            # Index
            df.index = [ii for _ in range(len(df))]
            # # Adds indices
            # df["index"] = ii
            # # Sorts columns
            # df = mngs.gen.mv_col(df, "index", 0)

            # Exclude edge effects
            edge_s = 1 / low_hz * 3
            indi_drop = (df.start_s < edge_s) + (
                xx.shape[-1] / fs - edge_s < df.end_s
            )
            df = df[~indi_drop]

            # Incidence [Hz]
            n_ripples = len(df)
            rec_s = xx.shape[-1] / fs
            df["incidence_hz"] = n_ripples / rec_s

            # Sorting
            sorted_columns = [
                "start_s",
                "end_s",
                "duration_s",
                "peak_s",
                "relative_peak_pos",
                "peak_amp_sd",
                "incidence_hz",
            ]
            df = df[sorted_columns]

            # Summarize as a dataframe across batch
            dfs.append(df)

        return pd.concat(dfs)  # .set_index("index")

    except ValueError as e:
        print("Caught an error:", e)
    # except Exception as e:
    #     print("Something wrong with ripple detection. :", e)


def main():
    xx, tt, fs = mngs.dsp.demo_sig(sig_type="ripple")
    df = detect_ripples(xx, fs)
    print(df)


if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF


# # append iEEG traces
# iEEG_traces = []
# for (
#     i_rip,
#     rip,
# ) in (
#     rip_df.reset_index().iterrows()
# ):  # rip_df = rips_df.iloc[i_trial]

#     start_pts = int(rip["start_time"] * CONFIG["FS_iEEG"])
#     end_pts = int(rip["end_time"] * CONFIG["FS_iEEG"])
#     # start_pts = int(rip["start_time"] * FS_TGT)
#     # end_pts = int(rip["end_time"] * FS_TGT)
#     iEEG_traces.append(
#         # iEEG[i_trial][:, start_pts:end_pts]
#         np.array(iEEG)[i_trial][:, start_pts:end_pts]
#     )
# rip_df["iEEG trace"] = iEEG_traces


# # append ripple band filtered iEEG traces
# ripple_band_iEEG_traces = []
# for (
#     i_rip,
#     rip,
# ) in (
#     rip_df.reset_index().iterrows()
# ):  # rip_df = rips_df.iloc[i_trial]
#     start_pts = int(rip["start_time"] * CONFIG["FS_iEEG"])
#     end_pts = int(rip["end_time"] * CONFIG["FS_iEEG"])

#     ripple_band_iEEG_traces.append(
#         iEEG_ripple_band_passed[i_trial][:, start_pts:end_pts]
#     )
# rip_df["ripple band iEEG trace"] = ripple_band_iEEG_traces

# # ripple peak amplitude
# ripple_peak_amplitude = [
#     np.abs(rbt).max(axis=-1) for rbt in ripple_band_iEEG_traces
# ]
# ripple_band_baseline_sd = iEEG_ripple_band_passed[i_trial].std(
#     axis=-1
# )
# rip_df["ripple_peak_amplitude_sd"] = [
#     (rpa / ripple_band_baseline_sd).mean()
#     for rpa in ripple_peak_amplitude
# ]

# rip_df["ripple_amplitude_sd"] = [
#     (np.abs(rbt).mean(axis=-1) / ripple_band_baseline_sd).mean()
#     for rbt in rip_df["ripple band iEEG trace"]
# ]



#!/usr/bin/env python3

import json
import os
import pickle
import warnings
from glob import glob

import h5py
import joblib
import mne
import mngs
import numpy as np
import pandas as pd
import torch
import yaml
from PIL import Image



def load(lpath, show=False, verbose=False, **kwargs):
    try:
        extension = "." + lpath.split(".")[-1]  # [REVISED]

        # csv
        if extension == ".csv":
            index_col = kwargs.get("index_col", 0)
            obj = pd.read_csv(lpath, **kwargs)
            obj = obj.loc[:, ~obj.columns.str.contains("^Unnamed")]
        # tsv
        elif extension == ".tsv":
            obj = pd.read_csv(lpath, sep="\t", **kwargs)
        # excel
        elif extension in [".xls", ".xlsx", ".xlsm", ".xlsb"]:
            obj = pd.read_excel(lpath, **kwargs)
        # parquet
        elif extension == ".parquet":
            obj = pd.read_parquet(lpath, **kwargs)

        # numpy
        elif extension == ".npy":
            obj = np.load(lpath, allow_pickle=True, **kwargs)

        # numpy npz
        elif extension == ".npz":
            obj = np.load(lpath)
            obj = dict(obj)
            obj = [v for v in obj.values()]

        # pkl
        elif extension == ".pkl":
            with open(lpath, "rb") as l:
                obj = pickle.load(l, **kwargs)
        # joblib
        elif extension == ".joblib":
            with open(lpath, "rb") as l:
                obj = joblib.load(l, **kwargs)
        # hdf5
        elif extension == ".hdf5":
            obj = {}
            with h5py.File(lpath, "r") as hf:
                for name in hf:  # [REVISED]
                    obj[name] = hf[name][:]
        # json
        elif extension == ".json":
            with open(lpath, "r") as f:
                obj = json.load(f)

        elif extension in [".jpg", ".png", ".tiff", "tif"]:
            obj = Image.open(lpath)
        # # png
        # elif extension == ".png":
        #     pass
        # # tiff
        # elif extension in [".tiff", ".tif"]:
        #     pass
        # yaml
        elif extension == ".yaml":

            lower = kwargs.pop("lower", False)

            with open(lpath) as f:
                obj = yaml.safe_load(f, **kwargs)  # [REVISED]

            if lower:
                obj = {k.lower(): v for k, v in obj.items()}

        # txt
        elif extension in [".txt", ".log", ".event"]:
            with open(lpath, "r") as f:  # [REVISED]
                obj = f.read().splitlines()  # [REVISED]

        # md
        elif extension == ".md":
            obj = load_markdown(lpath, **kwargs)
            
        # pth
        elif extension in [".pth", ".pt"]:
            obj = torch.load(lpath, **kwargs)

        # mat
        elif extension == ".mat":  # [REVISED]
            from pymatreader import read_mat  # [REVISED]

            obj = read_mat(lpath, **kwargs)  # [REVISED]
        # xml
        elif extension == ".xml":  # [REVISED]
            from xml2dict import xml2dict  # [REVISED]

            obj = xml2dict(lpath, **kwargs)  # [REVISED]
        # # edf
        # elif extension == ".edf":  # [REVISED]
        #     obj = mne.io.read_raw_edf(lpath, preload=True)  # [REVISED]
        # con
        elif extension == ".con":  # [REVISED]
            obj = mne.io.read_raw_fif(
                lpath, preload=True, **kwargs
            )  # [REVISED]
            obj = obj.to_data_frame()  # [REVISED]
            obj["samp_rate"] = obj.info["sfreq"]  # [REVISED]
        # # mrk
        # elif extension == ".mrk":  # [REVISED]
        #     obj = mne.io.read_mrk(lpath)  # [REVISED]

        # catboost model
        elif extension == ".cbm":  # [REVISED]
            from catboost import CatBoostModel  # [REVISED]

            obj = CatBoostModel.load_model(lpath, **kwargs)  # [REVISED]

        # EEG data
        elif extension in [
            ".vhdr",
            ".vmrk",
            ".edf",
            ".bdf",
            ".gdf",
            ".cnt",
            ".egi",
            ".eeg",
            ".set",
        ]:
            obj = _load_eeg_data(lpath, **kwargs)

        else:
            print(f"\nNot loaded from: {lpath}\n")
            return None

        if show or verbose:
            print(f"\nLoaded from: {lpath}\n")

        return obj

    except Exception as e:
        mngs.gen.print_block(f"{lpath} was not loaded:\n{e}", "red")
        # print(f"\n{lpath} was not loaded:\n{e}")

        # logging.error(f"\n{lpath} was not loaded:\n{e}")
        # return None


def _load_eeg_data(filename, **kwargs):
    """
    Load EEG data based on file extension and associated files using MNE-Python.

    Parameters:
    filename (str: The path to the file to be loaded.

    Returns:
    raw (mne.io.Raw: The loaded raw EEG data.
    """
    # Get the file extension
    extension = filename.split(".")[-1]

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)

        # Load the data based on the file extension
        if extension in ["vhdr", "vmrk"]:
            # Load BrainVision data
            raw = mne.io.read_raw_brainvision(filename, preload=True, **kwargs)
        elif extension == "edf":
            # Load European data format
            raw = mne.io.read_raw_edf(filename, preload=True, **kwargs)
        elif extension == "bdf":
            # Load BioSemi data format
            raw = mne.io.read_raw_bdf(filename, preload=True, **kwargs)
        elif extension == "gdf":
            # Load General data format
            raw = mne.io.read_raw_gdf(filename, preload=True, **kwargs)
        elif extension == "cnt":
            # Load Neuroscan CNT data
            raw = mne.io.read_raw_cnt(filename, preload=True, **kwargs)
        elif extension == "egi":
            # Load EGI simple binary data
            raw = mne.io.read_raw_egi(filename, preload=True, **kwargs)
        elif extension == "set":
            # ???
            raw = mne.io.read_raw(filename, preload=True, **kwargs)
        elif extension == "eeg":
            is_BrainVision = any(
                os.path.isfile(filename.replace(".eeg", ext))
                for ext in [".vhdr", ".vmrk"]
            )
            is_NihonKoden = any(
                os.path.isfile(filename.replace(".eeg", ext))
                for ext in [".21e", ".pnt", ".log"]
            )

            # Brain Vision
            if is_BrainVision:
                filename_v = filename.replace(".eeg", ".vhdr")
                raw = mne.io.read_raw_brainvision(
                    filename_v, preload=True, **kwargs
                )
            # Nihon Koden
            if is_NihonKoden:
                # raw = mne.io.read_raw_nihon(filename, preload=True, **kwargs)
                raw = mne.io.read_raw(filename, preload=True, **kwargs)
        else:
            raise ValueError(f"Unsupported file extension: {extension}")

        return raw


def load_markdown(lpath_md, style="plain_text"):
    import markdown
    import html2text
    
    # Load Markdown content from a file
    with open(lpath_md, 'r') as file:
        markdown_content = file.read()

    # Convert Markdown to HTML
    html_content = markdown.markdown(markdown_content)
    if style == "html":
        return html_content

    elif style == "plain_text":
        text_maker = html2text.HTML2Text()
        text_maker.ignore_links = True
        text_maker.bypass_tables = False
        plain_text = text_maker.handle(html_content)

        return plain_text

# def _check_encoding(file_path):
#     from chardet.universaldetector import UniversalDetector

#     detector = UniversalDetector()
#     with open(file_path, mode="rb") as f:
#         for binary in f:
#             detector.feed(binary)
#             if detector.done:
#                 break
#     detector.close()
#     enc = detector.result["encoding"]
#     return enc


# def get_data_path_from_a_package(package_str, resource):
#     import importlib
#     import os
#     import sys

#     spec = importlib.util.find_spec(package_str)
#     data_dir = os.path.join(spec.origin.split("src")[0], "data")
#     resource_path = os.path.join(data_dir, resource)
#     return resource_path


def load_yaml_as_an_optuna_dict(fpath_yaml, trial):
    _d = load(fpath_yaml)

    for k, v in _d.items():

        dist = v["distribution"]

        if dist == "categorical":
            _d[k] = trial.suggest_categorical(k, v["values"])

        elif dist == "uniform":
            _d[k] = trial.suggest_int(k, float(v["min"]), float(v["max"]))

        elif dist == "loguniform":
            _d[k] = trial.suggest_loguniform(
                k, float(v["min"]), float(v["max"])
            )

        elif dist == "intloguniform":
            _d[k] = trial.suggest_int(
                k, float(v["min"]), float(v["max"]), log=True
            )

    return _d


def load_study_rdb(study_name, rdb_raw_bytes_url):
    """
    study = load_study_rdb(
        study_name="YOUR_STUDY_NAME",
        rdb_raw_bytes_url="sqlite:///*.db"
    )
    """
    import optuna

    # rdb_raw_bytes_url = "sqlite:////tmp/fake/ywatanabe/_MicroNN_WindowSize-1.0-sec_MaxEpochs_100_2021-1216-1844/optuna_study_test_file#0.db"
    storage = optuna.storages.RDBStorage(url=rdb_raw_bytes_url)
    study = optuna.load_study(study_name=study_name, storage=storage)
    print(f"\nLoaded: {rdb_raw_bytes_url}\n")
    return study


def load_configs(IS_DEBUG=None, show=False, verbose=False):

    if os.getenv("CI") == "True":
        IS_DEBUG = True

    def update_debug(config, IS_DEBUG):
        if IS_DEBUG:
            debug_keys = mngs.gen.search("^DEBUG_", list(config.keys()))[1]
            for dk in debug_keys:
                dk_wo_debug_prefix = dk.split("DEBUG_")[1]
                config[dk_wo_debug_prefix] = config[dk]
                if show or verbose:
                    print(f"\n{dk} -> {dk_wo_debug_prefix}\n")
        return config

    # Check ./config/IS_DEBUG.yaml file if IS_DEBUG argument is not passed
    if IS_DEBUG is None:
        IS_DEBUG_PATH = "./config/IS_DEBUG.yaml"
        if os.path.exists(IS_DEBUG_PATH):
            IS_DEBUG = mngs.io.load("./config/IS_DEBUG.yaml").get("IS_DEBUG")
        else:
            IS_DEBUG = False

    # Main
    CONFIGS = {}
    for lpath in glob("./config/*.yaml"):
        CONFIG = update_debug(mngs.io.load(lpath), IS_DEBUG)
        CONFIGS.update(CONFIG)

    return CONFIGS



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-03 14:45:57 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/plt/ax/_raster.py


"""
This script does XYZ.
"""


"""
Imports
"""
import os
import re
import sys

import matplotlib
import matplotlib.pyplot as plt
import mngs
import seaborn as sns

mngs.gen.reload(mngs)
import warnings
from bisect import bisect_left
from glob import glob
from pprint import pprint

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
from icecream import ic
from natsort import natsorted
from tqdm import tqdm

"""
Warnings
"""
warnings.simplefilter("ignore", UserWarning)


"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""


def raster(ax, positions, time=None, **kwargs):
    def positions_to_df(positions, time):
        if time is None:
            time = np.arange(0, np.hstack(positions).max(), 1000)

        digi = np.nan * np.zeros([len(positions), len(time)], dtype=int)

        for i_unit, positions_unit in enumerate(positions):

            for pu in positions_unit:
                i_ins = bisect_left(time, pu)
                if i_ins == digi.shape[-1]:
                    i_ins -= i_unit
                digi[i_unit, i_ins] = i_unit

        return pd.DataFrame(np.array(digi)).T.set_index(time)

    df = positions_to_df(positions, time)

    # Use eventplot to create a raster plot
    ax.eventplot(positions, orientation="horizontal", **kwargs)

    # Set labels for axes
    ax.set_xlabel("Time Index")
    ax.set_ylabel("Channel Index")

    # Set y-ticks to correspond to each channel
    ax.set_yticks(range(len(df.columns)))
    ax.set_yticklabels(df.columns)

    return ax, df


def main():
    pass


if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



import pandas as pd
import torch
import numpy as np

def _format_a_sample_for_sktime(x):
    """
    x.shape: (n_chs, seq_len)
    """
    dims = pd.Series(
        [pd.Series(x[d], name=f"dim_{d}") for d in range(len(x))],
        index=[f"dim_{i}" for i in np.arange(len(x))]
    )
    return dims

def format_samples_for_sktime(X):
    """
    X.shape: (n_samples, n_chs, seq_len)
    """
    if torch.is_tensor(X):
        X = X.numpy() # (64, 160, 1024)

        X = X.astype(np.float64)

    return pd.DataFrame(
        [_format_a_sample_for_sktime(X[i]) for i in range(len(X))]
        )



#!/usr/bin/env python3

from ._get_proc_usages import get_proc_usages
from ._get_specs import get_specs
from ._rec_procs import rec_procs
from . import _utils

_ = None  # keep the importing orders

# from .limit_RAM import get_RAM, limit_RAM

# from .get import system_info

## EOF



#!/usr/bin/env python3

from torch.utils.data import Dataset
import numpy as np

class DefaultDataset(Dataset):
    """
    Apply transform for the first element of arrs_list

    Example:
        n = 1024
        n_chs = 19
        X = np.random.rand(n, n_chs, 1000)
        T = np.random.randint(0, 4, size=(n, 1))
        S = np.random.randint(0, 999, size=(n, 1))
        Sr = np.random.randint(0, 4, size=(n, 1))

        arrs_list = [X, T, S, Sr]
        transform = None
        ds = _DefaultDataset(arrs_list, transform=transform)
        len(ds) # 1024
    """

    def __init__(self, arrs_list, transform=None):
        self.arrs_list = arrs_list
        self.arrs = arrs_list # alias

        assert np.all([len(arr) for arr in arrs_list])

        self.length = len(arrs_list[0])
        self.transform = transform

    def __len__(self):
        return self.length

    def __getitem__(self, idx):
        arrs_list_idx = [arr[idx] for arr in self.arrs_list]

        # Here, you might want to transform, or apply DA on X as a numpy array
        if self.transform:
            dtype_orig = arrs_list_idx[0].dtype
            arrs_list_idx[0] = self.transform(arrs_list_idx[0].astype(np.float64))\
                                   .astype(dtype_orig)
        return arrs_list_idx



#!/usr/bin/env python3

from ._bonferroni_correction import bonferroni_correction
from ._brunner_munzel_test import brunner_munzel_test
from ._calc_partial_corr import calc_partial_corr
from ._corr_test import corr_test
from ._fdr_correction import fdr_correction, fdr_correction_torch
from ._general import kurtosis, mean, median, q, skewness, std, zscore
from ._multicompair import multicompair
from ._nocorrelation_test import nocorrelation_test
from ._smirnov_grubbs import smirnov_grubbs
from ._to_asterisks import to_asterisks



#!/usr/bin/env python3

import warnings
from itertools import cycle

import numpy as np
from sklearn.metrics import average_precision_score, precision_recall_curve


def solve_the_intersection_of_a_line_and_iso_f1_curve(f1, a, b):
    """
    Determines the intersection of the following lines:
        1) a line: y = a * x + b
        2) the iso-f1 curve: y = f1 * x / (2 * x - f1)
    , where a, b, and f1 are the constant values.
    """
    _a = 2 * a
    _b = -a * f1 + 2 * b - f1
    _c = -b * f1

    x_f = (-_b + np.sqrt(_b**2 - 4 * _a * _c)) / (2 * _a)
    y_f = a * x_f + b

    return (x_f, y_f)


def to_onehot(labels, n_classes):
    eye = np.eye(n_classes, dtype=int)
    return eye[labels]


def pre_rec_auc(plt, true_class, pred_proba, labels):
    """
    Calculates the precision recall curve.
    """

    # Use label_binarize to be multi-label like settings
    n_classes = len(labels)
    true_class_onehot = to_onehot(true_class, n_classes)

    # For each class
    precision = dict()
    recall = dict()
    threshold = dict()
    pre_rec_auc = dict()
    for i in range(n_classes):
        true_class_i_onehot = true_class_onehot[:, i]
        pred_proba_i = pred_proba[:, i]

        try:
            precision[i], recall[i], threshold[i] = precision_recall_curve(
                true_class_i_onehot,
                pred_proba_i,
            )
            pre_rec_auc[i] = average_precision_score(true_class_i_onehot, pred_proba_i)
        except Exception as e:
            print(e)
            precision[i], recall[i], threshold[i], pre_rec_auc[i] = (
                np.nan,
                np.nan,
                np.nan,
                np.nan,
            )

    ## Average precision: micro and macro

    # A "micro-average": quantifying score on all classes jointly
    precision["micro"], recall["micro"], threshold["micro"] = precision_recall_curve(
        true_class_onehot.ravel(), pred_proba.ravel()
    )
    pre_rec_auc["micro"] = average_precision_score(
        true_class_onehot, pred_proba, average="micro"
    )

    # macro
    _pre_rec_aucs = []
    for i in range(n_classes):
        try:
            _pre_rec_aucs.append(
                average_precision_score(
                    true_class_onehot[:, i], pred_proba[:, i], average="macro"
                )
            )
        except Exception as e:
            print(
                f'\nPRE-REC-AUC for "{labels[i]}" was not defined and NaN-filled '
                "for a calculation purpose (for the macro avg.)\n"
            )
            _pre_rec_aucs.append(np.nan)
    pre_rec_auc["macro"] = np.nanmean(_pre_rec_aucs)

    # pre_rec_auc["macro"] = average_precision_score(
    #     true_class_onehot, pred_proba, average="macro"
    # )

    # Plot Precision-Recall curve for each class and iso-f1 curves
    colors = cycle(["navy", "turquoise", "darkorange", "cornflowerblue", "teal"])
    fig, ax = plt.subplots()
    ax.set_box_aspect(1)
    lines = []
    legends = []

    # iso-F1: By definition, an iso-F1 curve contains all points
    #         in the precision/recall space whose F1 scores are the same.
    f_scores = np.linspace(0.2, 0.8, num=4)
    # for f_score in f_scores:
    for i_f, f_score in enumerate(f_scores):
        x = np.linspace(0.01, 1)  # num=50
        y = f_score * x / (2 * x - f_score)
        (l,) = ax.plot(x[y >= 0], y[y >= 0], color="gray", alpha=0.2)

        # ax.annotate("f1={0:0.1f}".format(f_score), xy=(0.9, y[45] + 0.02))
        x_f, y_f = solve_the_intersection_of_a_line_and_iso_f1_curve(f_score, 0.5, 0.5)
        ax.annotate("f1={0:0.1f}".format(f_score), xy=(x_f - 0.1, y_f - 0.1 * 0.5))
        # ax.annotate("f1={0:0.1f}".format(f_score), xy=(y[35] - 0.02 * (3 - i_f), 0.85))

    lines.append(l)
    legends.append("iso-f1 curves")

    """
    ## In this project, average precision-recall curve is not drawn.
    (l,) = ax.plot(recall["micro"], precision["micro"], color="gold", lw=2)
    lines.append(l)
    legends.append("micro-average\n(AUC = {0:0.2f})" "".format(pre_rec_auc["micro"]))
    """

    ## Each Class
    for i, color in zip(range(n_classes), colors):
        (l,) = ax.plot(recall[i], precision[i], color=color, lw=2)
        lines.append(l)
        legends.append("{0} (AUC = {1:0.2f})" "".format(labels[i], pre_rec_auc[i]))

    # fig = plt.gcf()
    fig.subplots_adjust(bottom=0.25)
    ax.set_xlim([-0.01, 1.01])
    ax.set_ylim([-0.01, 1.01])
    ax.set_xticks([0.0, 0.5, 1.0])
    ax.set_yticks([0.0, 0.5, 1.0])
    ax.set_xlabel("Recall")
    ax.set_ylabel("Precision")
    ax.set_title("Precision-Recall Curve")
    ax.legend(lines, legends, loc="lower left")

    metrics = dict(
        pre_rec_auc=pre_rec_auc,
        precision=precision,
        recall=recall,
        threshold=threshold,
    )

    return fig, metrics


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np
    from scipy.special import softmax
    from sklearn import datasets, svm
    from sklearn.model_selection import train_test_split

    def mk_demo_data(n_classes=2, batch_size=16):
        labels = ["cls{}".format(i_cls) for i_cls in range(n_classes)]
        true_class = np.random.randint(0, n_classes, size=(batch_size,))
        pred_proba = softmax(np.random.rand(batch_size, n_classes), axis=-1)
        pred_class = np.argmax(pred_proba, axis=-1)
        return labels, true_class, pred_proba, pred_class

    ## Fix seed
    np.random.seed(42)

    """
    ################################################################################
    ## A Minimal Example
    ################################################################################    
    labels, true_class, pred_proba, pred_class = \
        mk_demo_data(n_classes=10, batch_size=256)

    pre_rec_auc, precision, recall, threshold = \
        calc_pre_rec_auc(true_class, pred_proba, labels, plot=False)
    """

    ################################################################################
    ## MNIST
    ################################################################################
    from sklearn import datasets, metrics, svm
    from sklearn.model_selection import train_test_split

    digits = datasets.load_digits()

    # flatten the images
    n_samples = len(digits.images)
    data = digits.images.reshape((n_samples, -1))

    # Create a classifier: a support vector classifier
    clf = svm.SVC(gamma=0.001, probability=True)

    # Split data into 50% train and 50% test subsets
    X_train, X_test, y_train, y_test = train_test_split(
        data, digits.target, test_size=0.5, shuffle=False
    )

    # Learn the digits on the train subset
    clf.fit(X_train, y_train)

    # Predict the value of the digit on the test subset
    predicted_proba = clf.predict_proba(X_test)
    predicted = clf.predict(X_test)

    n_classes = len(np.unique(digits.target))
    labels = ["Class {}".format(i) for i in range(n_classes)]

    ## Configures matplotlib
    plt.rcParams["font.size"] = 20
    plt.rcParams["legend.fontsize"] = "xx-small"
    plt.rcParams["figure.figsize"] = (16 * 1.2, 9 * 1.2)

    ## Main
    fig, metrics_dict = pre_rec_auc(plt, y_test, predicted_proba, labels)

    fig.show()

    print(metrics_dict.keys())
    # dict_keys(['pre_rec_auc', 'precision', 'recall', 'threshold'])



#!/usr/bin/env python3


import matplotlib


class OOMFormatter(matplotlib.ticker.ScalarFormatter):
    # https://stackoverflow.com/questions/42656139/set-scientific-notation-with-fixed-exponent-and-significant-digits-for-multiple
    # def __init__(self, order=0, fformat="%1.1f", offset=True, mathText=True):
    def __init__(self, order=0, fformat="%1.0d", offset=True, mathText=True):
        self.oom = order
        self.fformat = fformat
        matplotlib.ticker.ScalarFormatter.__init__(
            self, useOffset=offset, useMathText=mathText
        )

    def _set_order_of_magnitude(self):
        self.orderOfMagnitude = self.oom

    def _set_format(self, vmin=None, vmax=None):
        self.format = self.fformat
        if self._useMathText:
            self.format = r"$\mathdefault{%s}$" % self.format


def sci_note(
    ax,
    order,
    fformat="%1.0d",
    x=False,
    y=False,
    scilimits=(-3, 3),
):
    """
    Change the expression of the x- or y-axis to the scientific notation like *10^3
    , where 3 is the first argument, order.

    Example:
        order = 4 # 10^4
        ax = sci_note(
                 ax,
                 order,
                 fformat="%1.0d",
                 x=True,
                 y=False,
                 scilimits=(-3, 3),
    """

    if x == True:
        ax.xaxis.set_major_formatter(
            OOMFormatter(order=order, fformat=fformat)
        )
        ax.ticklabel_format(axis="x", style="sci", scilimits=scilimits)
    if y == True:
        ax.yaxis.set_major_formatter(
            OOMFormatter(order=order, fformat=fformat)
        )
        ax.ticklabel_format(axis="y", style="sci", scilimits=scilimits)

    return ax



#!/usr/bin/env python3
# Time-stamp: "2021-11-27 18:45:23 (ylab)"

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt


def mk_patches(colors, labels):
    """
    colors = ["red", "blue"]
    labels = ["label_1", "label_2"]
    ax.legend(handles=mngs.plt.mk_patches(colors, labels))
    """

    patches = [mpatches.Patch(color=c, label=l) for c, l in zip(colors, labels)]
    return patches



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: 2024-05-12 21:02:21 (7)
# /sshx:ywatanabe@444:/home/ywatanabe/proj/mngs/src/mngs/general/_title2spath.py


def title2path(title):
    path = title

    # Comma patterns
    patterns = [":", ";", "=", "[", "]"]
    for pp in patterns:
        path = path.replace(pp, "")

    # Exceptions
    path = path.replace("_-_", "-")
    path = path.replace(" ", "_")

    # Consective under scores
    for _ in range(10):
        path = path.replace("__", "_")

    return path.lower()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-29 11:03:51 (ywatanabe)"

import sys

import matplotlib.pyplot as plt
import mngs
import numpy as np
from mngs.general import batch_fn, torch_fn
from mngs.nn import PAC

"""
mngs.dsp.pac function
"""


@torch_fn
@batch_fn
def pac(
    x,
    fs,
    pha_start_hz=2,
    pha_end_hz=20,
    pha_n_bands=100,
    amp_start_hz=60,
    amp_end_hz=160,
    amp_n_bands=100,
    device="cuda",
    batch_size=1,
    fp16=False,
    trainable=False,
    n_perm=None,
    amp_prob=False,
):
    """
    Compute the phase-amplitude coupling (PAC) for signals. This function automatically handles inputs as
    PyTorch tensors, NumPy arrays, or pandas DataFrames.

    Arguments:
    - x (torch.Tensor | np.ndarray | pd.DataFrame): Input signal. Shape can be either (batch_size, n_chs, seq_len) or
    - fs (float): Sampling frequency of the input signal.
    - pha_start_hz (float, optional): Start frequency for phase bands. Default is 2 Hz.
    - pha_end_hz (float, optional): End frequency for phase bands. Default is 20 Hz.
    - pha_n_bands (int, optional): Number of phase bands. Default is 100.
    - amp_start_hz (float, optional): Start frequency for amplitude bands. Default is 60 Hz.
    - amp_end_hz (float, optional): End frequency for amplitude bands. Default is 160 Hz.
    - amp_n_bands (int, optional): Number of amplitude bands. Default is 100.

    Returns:
    - torch.Tensor: PAC values. Shape: (batch_size, n_chs, pha_n_bands, amp_n_bands)
    - numpy.ndarray: Phase bands used for the computation.
    - numpy.ndarray: Amplitude bands used for the computation.

    Example:
        FS = 512
        T_SEC = 4
        xx, tt, fs = mngs.dsp.demo_sig(
            batch_size=1, n_chs=1, fs=FS, t_sec=T_SEC, sig_type="tensorpac"
        )
        pac, pha_mids_hz, amp_mids_hz = mngs.dsp.pac(xx, fs)
    """
    m = PAC(
        x.shape[-1],
        fs,
        pha_start_hz=pha_start_hz,
        pha_end_hz=pha_end_hz,
        pha_n_bands=pha_n_bands,
        amp_start_hz=amp_start_hz,
        amp_end_hz=amp_end_hz,
        amp_n_bands=amp_n_bands,
        fp16=fp16,
        trainable=trainable,
        n_perm=n_perm,
        amp_prob=amp_prob,
    ).to(device)
    return m(x.to(device)), m.PHA_MIDS_HZ, m.AMP_MIDS_HZ


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import mngs
    import numpy as np

    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(sys, plt)

    # Parameters
    FS = 512
    T_SEC = 4
    # IS_TRAINABLE = False
    # FP16 = True

    for IS_TRAINABLE in [True, False]:
        for FP16 in [True, False]:

            # Demo signal
            xx, tt, fs = mngs.dsp.demo_sig(
                batch_size=1, n_chs=1, fs=FS, t_sec=T_SEC, sig_type="pac"
            )

            # mngs calculation
            pac_mngs, pha_mids_mngs, amp_mids_mngs = mngs.dsp.pac(
                xx,
                fs,
                pha_n_bands=50,
                amp_n_bands=30,
                trainable=IS_TRAINABLE,
                fp16=FP16,
            )
            i_batch, i_ch = 0, 0
            pac_mngs = pac_mngs[i_batch, i_ch]

            # Tensorpac calculation
            (
                _,
                _,
                _pha_mids_tp,
                _amp_mids_tp,
                pac_tp,
            ) = mngs.dsp.utils.pac.calc_pac_with_tensorpac(xx, fs, T_SEC)

            # Validates the consitency in frequency definitions
            assert np.allclose(pha_mids_mngs, _pha_mids_tp)
            assert np.allclose(amp_mids_mngs, _amp_mids_tp)

            mngs.io.save(
                (pac_mngs, pac_tp, pha_mids_mngs, amp_mids_mngs),
                "./data/cache.npz",
            )

            # ################################################################################
            # # cache
            # pac_mngs, pac_tp, pha_mids_mngs, amp_mids_mngs = mngs.io.load(
            #     "./data/cache.npz"
            # )
            # ################################################################################

            # Plots
            fig = mngs.dsp.utils.pac.plot_PAC_mngs_vs_tensorpac(
                pac_mngs, pac_tp, pha_mids_mngs, amp_mids_mngs
            )
            fig.suptitle(
                "Phase-Amplitude Coupling calculation\n\n(Bandpass Filtering -> Hilbert Transformation-> Modulation Index)"
            )
            plt.show()

            mngs.gen.reload(mngs.dsp)

            # Saves the figure
            trainable_str = "trainable" if IS_TRAINABLE else "static"
            fp_str = "fp16" if FP16 else "fp32"
            mngs.io.save(
                fig, f"pac_with_{trainable_str}_bandpass_{fp_str}.png"
            )

# EOF

"""
/home/ywatanabe/proj/entrance/mngs/dsp/_pac.py
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-01-24 13:56:36 (ywatanabe)"

import random


def sliding_window_data_augmentation(x, window_size_pts):
    start = random.randint(0, x.shape[-1] - window_size_pts)
    end = start + window_size_pts
    return x[..., start:end]



# import torch.nn as nn
import torch.optim as optim
from .Ranger_Deep_Learning_Optimizer.ranger.ranger2020 import Ranger


# def set_an_optim(models, optim_str, lr):
def set(models, optim_str, lr):    
    """Sets an optimizer to models"""
    if not isinstance(models, list):
        models = [models]
    learnable_params = []
    for m in models:
        learnable_params += list(m.parameters())
    # optim = mngs.ml.switch_optim(optim_str)
    optim = get(optim_str)    
    return optim(learnable_params, lr)

def get(optim_str):
    optims_dict = {
        "adam": optim.Adam,
        "ranger": Ranger,
        "rmsprop": optim.RMSprop,
        "sgd": optim.SGD
        }
    return optims_dict[optim_str]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-05-30 11:06:27 (ywatanabe)"

"""
mngs.dsp.wavelet function
"""

import mngs
from mngs.general import batch_fn, torch_fn
from mngs.nn import Wavelet


# Functions
@torch_fn
@batch_fn
def wavelet(
    x,
    fs,
    freq_scale="linear",
    out_scale="linear",
    device="cuda",
    batch_size=32,
):
    m = (
        Wavelet(fs, freq_scale=freq_scale, out_scale="linear")
        .to(device)
        .eval()
    )
    pha, amp, freqs = m(x.to(device))

    if out_scale == "log":
        amp = (amp + 1e-5).log()
        if amp.isnan().any():
            print("NaN is detected while taking the lograrithm of amplitude.")

    return pha, amp, freqs


# @torch_fn
# def wavelet(
#     x,
#     fs,
#     freq_scale="linear",
#     out_scale="linear",
#     device="cuda",
#     batch_size=32,
# ):
#     @torch_fn
#     def _wavelet(
#         x,
#         fs,
#         freq_scale="linear",
#         out_scale="linear",
#         device="cuda",
#     ):
#         m = (
#             Wavelet(fs, freq_scale=freq_scale, out_scale=out_scale)
#             .to(device)
#             .eval()
#         )
#         pha, amp, freqs = m(x.to(device))

#         if out_scale == "log":
#             amp = (amp + 1e-5).log()
#             if amp.isnan().any():
#                 print(
#                     "NaN is detected while taking the lograrithm of amplitude."
#                 )

#         return pha, amp, freqs

#     if len(x) <= batch_size:
#         try:
#             pha, amp, freqs = _wavelet(
#                 x,
#                 fs,
#                 freq_scale=freq_scale,
#                 out_scale=out_scale,
#                 device=device,
#             )
#             torch.cuda.empty_cache()
#             return pha, amp, freqs

#         except Exception as e:
#             print(e)
#             print("\nTrying Batch Mode...")

#     n_batches = (len(x) + batch_size - 1) // batch_size
#     device_orig = x.device
#     pha, amp, freqs = [], [], []
#     for i_batch in tqdm(range(n_batches)):
#         start = i_batch * batch_size
#         end = (i_batch + 1) * batch_size
#         _pha, _amp, _freqs = _wavelet(
#             x[start:end],
#             fs,
#             freq_scale=freq_scale,
#             out_scale=out_scale,
#             device=device,
#         )
#         torch.cuda.empty_cache()
#         # to CPU
#         pha.append(_pha.cpu())
#         amp.append(_amp.cpu())
#         freqs.append(_freqs.cpu())

#     pha = torch.vstack(pha)
#     amp = torch.vstack(amp)
#     freqs = freqs[0]

#     try:
#         pha = pha.to(device_orig)
#         amp = amp.to(device_orig)
#         freqs = freqs.to(device_orig)
#     except Exception as e:
#         print(
#             f"\nError occurred while transferring wavelet outputs back to the original device. Proceeding with CPU tensor. \n\n({e})"
#         )

#     sleep(0.5)
#     torch.cuda.empty_cache()
#     return pha, amp, freqs


if __name__ == "__main__":
    import sys

    import matplotlib.pyplot as plt
    import numpy as np

    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, agg=True
    )

    # Parameters
    FS = 512
    SIG_TYPE = "chirp"
    T_SEC = 4

    # Demo signal
    xx, tt, fs = mngs.dsp.demo_sig(
        batch_size=64,
        n_chs=19,
        n_segments=2,
        t_sec=T_SEC,
        fs=FS,
        sig_type=SIG_TYPE,
    )

    if SIG_TYPE in ["tensorpac", "pac"]:
        i_segment = 0
        xx = xx[:, :, i_segment, :]

    # Main
    pha, amp, freqs = wavelet(xx, fs, device="cuda")
    freqs = freqs[0, 0]

    # Plots
    i_batch, i_ch = 0, 0
    fig, axes = mngs.plt.subplots(nrows=3)

    # # Time vector for x-axis extents
    # time_extent = [tt.min(), tt.max()]

    # Trace
    axes[0].plot(tt, xx[i_batch, i_ch], label=SIG_TYPE)
    axes[0].set_ylabel("Amplitude [?V]")
    axes[0].legend(loc="upper left")
    axes[0].set_title("Signal")

    # Amplitude
    # extent = [time_extent[0], time_extent[1], freqs.min(), freqs.max()]
    axes[1].imshow2d(
        np.log(amp[i_batch, i_ch] + 1e-5).T,
        cbar_label="Log(amplitude [?V]) [a.u.]",
        aspect="auto",
        # extent=extent,
        # origin="lower",
    )
    axes[1] = mngs.plt.ax.set_ticks(axes[1], x_ticks=tt, y_ticks=freqs)
    axes[1].set_ylabel("Frequency [Hz]")
    axes[1].set_title("Amplitude")

    # Phase
    axes[2].imshow2d(
        pha[i_batch, i_ch].T,
        cbar_label="Phase [rad]",
        aspect="auto",
        # extent=extent,
        # origin="lower",
    )
    axes[2] = mngs.plt.ax.set_ticks(axes[2], x_ticks=tt, y_ticks=freqs)
    axes[2].set_ylabel("Frequency [Hz]")
    axes[2].set_title("Phase")

    fig.suptitle("Wavelet Transformation")
    fig.supxlabel("Time [s]")

    for ax in axes:
        ax = mngs.plt.ax.set_n_ticks(ax)
        # ax.set_xlim(time_extent[0], time_extent[1])

    fig.tight_layout(rect=[0, 0.03, 1, 0.95])

    mngs.io.save(fig, "wavelet.png")

    # Close
    mngs.gen.close(CONFIG)

# EOF

"""
/home/ywatanabe/proj/entrance/mngs/dsp/_wavelet.py
"""



#!/usr/bin/env python3
# Time-stamp: "2021-09-25 15:39:51 (ylab)"

import numpy as np
import torch
import mngs


def bonferroni_correction(pval, alpha=0.05):
    # https://github.com/mne-tools/mne-python/blob/main/mne/stats/multi_comp.py
    """P-value correction with Bonferroni method.

    Parameters
    ----------
    pval : array_like
        Set of p-values of the individual tests.
    alpha : float
        Error rate.

    Returns
    -------
    reject : array, bool
        True if a hypothesis is rejected, False if not.
    pval_corrected : array
        P-values adjusted for multiple hypothesis testing to limit FDR.
    """
    pval = np.asarray(pval)
    pval_corrected = pval * float(pval.size)
    # p-values must not be larger than 1.
    pval_corrected = pval_corrected.clip(max=1.0)
    reject = pval_corrected < alpha
    return reject, pval_corrected


def bonferroni_correction_torch(pvals, alpha=0.05):
    """P-value correction with Bonferroni method.

    Parameters
    ----------
    pvals : array_like
        Set of p-values of the individual tests.
    alpha : float
        Error rate.

    Returns
    -------
    reject : array, bool
        True if a hypothesis is rejected, False if not.
    pvals_corrected : array
        P-values adjusted for multiple hypothesis testing to limit FDR.
    """
    pvals = torch.tensor(pvals)
    pvals_corrected = pvals * torch.tensor(pvals.size()).float()
    # p-values must not be larger than 1.
    pvals_corrected = pvals_corrected.clip(max=1.0)
    reject = pvals_corrected < alpha
    return reject, pvals_corrected


if __name__ == "__main__":
    pvals_npy = np.array([0.02, 0.03, 0.05])
    pvals_torch = torch.tensor(np.array([0.02, 0.03, 0.05]))

    reject, pvals_corrected = bonferroni_correction(pvals_npy, alpha=0.05)

    reject_torch, pvals_corrected_torch = bonferroni_correction_torch(
        pvals_torch, alpha=0.05
    )

    arr = pvals_corrected.astype(float)
    tor = pvals_corrected_torch.numpy().astype(float)
    print(mngs.general.isclose(arr, tor))



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-04 08:33:07 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/linalg/_distance.py


import scipy.spatial.distance as _distance
from mngs.gen import wrap


@wrap
def euclidean_distance(*args, **kwargs):
    return _distance.euclidean(*args, **kwargs)


@wrap
def cdist(*args, **kwargs):
    return _distance.cdist(*args, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-26 20:02:00 (ywatanabe)"


import numpy as np
import torch
from mngs.general import torch_fn

# def to_z(x, axis):
#     if isinstance(x, torch.Tensor):  # [REVISED]
#         return (x - x.mean(dim=axis, keepdim=True)) / x.std(
#             dim=axis, keepdim=True
#         )  # [REVISED]
#     if isinstance(x, np.ndarray):  # [REVISED]
#         return (x - x.mean(axis=axis, keepdims=True)) / x.std(
#             axis=axis, keepdims=True
#         )


@torch_fn
def to_z(x, dim=-1, device="cuda"):
    return (x - x.mean(dim=dim, keepdim=True)) / x.std(dim=dim, keepdim=True)


@torch_fn
def to_1_1(x, amp=1.0, dim=-1, fn="mean", device="cuda"):
    MM = x.max(dim=dim, keepdims=True)[0].abs()
    mm = x.min(dim=dim, keepdims=True)[0].abs()
    return amp * x / torch.maximum(MM, mm)


@torch_fn
def unbias(x, dim=-1, fn="mean", device="cuda"):
    if fn == "mean":
        return x - x.mean(dim=dim, keepdims=True)
    if fn == "min":
        return x - x.min(dim=dim, keepdims=True)[0]


# def to_z(x, axis):
#     if torch.Tensor:
#         return (x - x.mean(dim=axis, keepdims=True)) / x.std(dim=axis, keepdims=True)
#     if np.ndarray:
#         return (x - x.mean(axis=axis, keepdims=True)) / x.std(axis=axis, keepdims=True)

#     x, x_type = mngs.gen.my2array(x)

#     dtype_orig = x.dtype
#     x = x.astype(np.float64)
#     z = (x - x.mean(axis=axis, keepdims=True)) / x.std(axis=axis, keepdims=True)
#     z = z.astype(dtype_orig)

#     if x_type == "tensor":
#         return torch.tensor(z)
#     if x_type == "numpy":
#         return z



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-11 06:52:12 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/ml/_gen_AI/_ChatGPT.py


"""
This script does XYZ.
"""


"""
Imports
"""
import sys

import matplotlib.pyplot as plt
import mngs
from openai import OpenAI

from ._BaseGenAI import BaseGenAI

# sys.path = ["."] + sys.path
# from scripts import utils, load

"""
Warnings
"""
# warnings.simplefilter("ignore", UserWarning)


"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""


class ChatGPT(BaseGenAI):
    def __init__(
        self,
        system_setting="",
        model="",
        api_key="",
        stream=False,
        seed=None,
        n_keep=1,
        temperature=1.0,
    ):
        super().__init__(
            system_setting=system_setting,
            model=model,
            api_key=api_key,
            stream=stream,
            n_keep=n_keep,
            temperature=temperature,
        )

    def _init_client(
        self,
    ):
        client = OpenAI()
        return client

    def _api_call_static(self):
        out_text = (
            self.client.chat.completions.create(
                model=self.model,
                messages=self.chat_history,
                seed=self.seed,
                stream=False,
                temperature=self.temperature,
            )
            .choices[0]
            .message.content
        )
        return out_text

    def _api_call_stream(self):
        stream = self.client.chat.completions.create(
            model=self.model,
            messages=self.chat_history,
            max_tokens=4096,  # You can adjust this as needed
            n=1,
            stream=self.stream,
            seed=self.seed,
            temperature=self.temperature,
        )

        for chunk in stream:
            # if chunk.choices:
            current_text = chunk.choices[0].delta.content
            if current_text:
                yield f"{current_text}"

    def _get_available_models(self):
        return [m.id for m in OpenAI(api_key=self.api_key).models.list()]


def main():
    m = mngs.ai.GenAI("gpt-4o")
    m("Hi")
    pass


if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-04 07:24:04 (ywatanabe)"

from ._circular_hist import circular_hist
from ._extend import extend
from ._fill_between import fill_between, plot_with_ci
from ._force_aspect import force_aspect
from ._hide_spines import hide_spines
from ._imshow2d import imshow2d
from ._map_ticks import map_ticks
from ._panel import panel
from ._raster import raster
from ._rectangle import rectangle
from ._sci_note import sci_note
from ._set_n_ticks import set_n_ticks
from ._set_size import set_size
from ._set_supxyt import set_supxyt
from ._set_ticks import set_ticks
from ._set_xyt import set_xyt

# from ._set_pos import set_pos
from ._shift import shift



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-07 22:16:25 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/general/_not_implemented.py

import warnings


def not_implemented(func):
    """
    Decorator to mark methods as not implemented, issue a warning, and prevent their execution.

    Arguments:
        func (callable): The function or method to decorate.

    Returns:
        callable: A wrapper function that issues a warning and raises NotImplementedError when called.
    """

    def wrapper(*args, **kwargs):
        # Issue a warning before raising the error
        warnings.warn(
            f"Attempt to use unimplemented method: '{func.__name__}'. This method is not yet available.",
            category=FutureWarning,
            stacklevel=2,
        )
        # # Raise the NotImplementedError
        # raise NotImplementedError(f"The method '{func.__name__}' is not implemented yet.")

    return wrapper



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-05 01:23:04 (ywatanabe)"
# Author: Yusuke Watanabe (ywata1989@gmail.com)

from ._find import find_dir, find_file, find_git_root
from ._path import file_size, spath, split, this_path
from ._version import find_latest, increment_version



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-29 10:59:21 (ywatanabe)"

"""
This script defines the ModulationIndex module.
"""

# Imports
import sys

import mngs
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F


# Functions
class ModulationIndex(nn.Module):
    def __init__(self, n_bins=18, fp16=False, amp_prob=False):
        super(ModulationIndex, self).__init__()
        self.n_bins = n_bins
        self.fp16 = fp16
        self.register_buffer(
            "pha_bin_cutoffs", torch.linspace(-np.pi, np.pi, n_bins + 1)
        )

        self.amp_prob = amp_prob

    @property
    def pha_bin_centers(
        self,
    ):
        return (
            ((self.pha_bin_cutoffs[1:] + self.pha_bin_cutoffs[:-1]) / 2)
            .detach()
            .cpu()
            .numpy()
        )

    def forward(self, pha, amp, epsilon=1e-9):
        """
        Compute the Modulation Index based on phase (pha) and amplitude (amp) tensors.

        Parameters:
        - pha (torch.Tensor): Tensor of phase values with shape
                              (batch_size, n_channels, n_freqs_pha, n_segments, sequence_length).
        - amp (torch.Tensor): Tensor of amplitude values with a similar shape as pha.
                              (batch_size, n_channels, n_freqs_amp, n_segments, sequence_length).

        Returns:
        - MI (torch.Tensor): The Modulation Index for each batch and channel.
        """
        assert pha.ndim == amp.ndim == 5

        if self.fp16:
            pha, amp = pha.half().contiguous(), amp.half().contiguous()
        else:
            pha, amp = pha.float().contiguous(), amp.float().contiguous()

        device = pha.device

        pha_masks = self._phase_to_masks(pha, self.pha_bin_cutoffs.to(device))
        # (batch_size, n_channels, n_freqs_pha, n_segments, sequence_length, n_bins)

        # Expands amp and masks to utilize broadcasting
        # i_batch = 0
        # i_chs = 1
        i_freqs_pha = 2
        i_freqs_amp = 3
        # i_segments = 4
        i_time = 5
        i_bins = 6

        # Coupling
        pha_masks = pha_masks.unsqueeze(i_freqs_amp)
        amp = amp.unsqueeze(i_freqs_pha).unsqueeze(i_bins)

        amp_bins = pha_masks * amp  # this is the most memory-consuming process

        # # Batch processing to reduce maximum VRAM occupancy
        # pha_masks = self.dh_pha.fit(pha_masks, keepdims=[2, 3, 5, 6])
        # amp = self.dh_amp.fit(amp, keepdims=[2, 3, 5, 6])
        # n_chunks = len(pha_masks) // self.chunk_size
        # amp_bins = []
        # for i_chunk in range(n_chunks):
        #     start = i_chunk * self.chunk_size
        #     end = (i_chunk + 1) * self.chunk_size
        #     _amp_bins = pha_masks[start:end] * amp[start:end]
        #     amp_bins.append(_amp_bins.cpu())
        # amp_bins = torch.cat(amp_bins)
        # amp_bins = self.dh_pha.unfit(amp_bins)
        # pha_masks = self.dh_pha.unfit(pha_masks)
        # Takes mean amplitude in each bin
        amp_sums = amp_bins.sum(dim=i_time, keepdims=True).to(device)
        counts = pha_masks.sum(dim=i_time, keepdims=True)
        amp_means = amp_sums / (counts + epsilon)

        amp_probs = amp_means / (
            amp_means.sum(dim=-1, keepdims=True) + epsilon
        )

        if self.amp_prob:
            return amp_probs.detach().cpu()

        """
        matplotlib.use("TkAgg")
        fig, ax = mngs.plt.subplots(subplot_kw={'polar': True})
        yy = amp_probs[0, 0, 0, 0, 0, 0, :].detach().cpu().numpy()
        xx = ((self.pha_bin_cutoffs[1:] + self.pha_bin_cutoffs[:-1]) / 2).detach().cpu().numpy()
        ax.bar(xx, yy, width=.1)
        plt.show()
        """

        MI = (
            torch.log(torch.tensor(self.n_bins, device=device) + epsilon)
            + (amp_probs * (amp_probs + epsilon).log()).sum(dim=-1)
        ) / torch.log(torch.tensor(self.n_bins, device=device))

        # Squeeze the n_bin dimension
        MI = MI.squeeze(-1)

        # Takes mean along the n_segments dimension
        i_segment = -1
        MI = MI.mean(axis=i_segment)

        if MI.isnan().any():
            raise ValueError(
                "NaN values detected in Modulation Index calculation."
            )

        return MI

    @staticmethod
    def _phase_to_masks(pha, phase_bin_cutoffs):
        n_bins = int(len(phase_bin_cutoffs) - 1)
        bin_indices = (
            (torch.bucketize(pha, phase_bin_cutoffs, right=False) - 1).clamp(
                0, n_bins - 1
            )
        ).long()
        one_hot_masks = (
            F.one_hot(
                bin_indices,
                num_classes=n_bins,
            )
            .bool()
            .to(pha.device)
        )
        return one_hot_masks


def _reshape(x, batch_size=2, n_chs=4):
    return (
        torch.tensor(x)
        .float()
        .unsqueeze(0)
        .unsqueeze(0)
        .repeat(batch_size, n_chs, 1, 1, 1)
    )


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import mngs

    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, fig_scale=3
    )

    # Parameters
    FS = 512
    T_SEC = 1
    device = "cuda"

    # Demo signal
    xx, tt, fs = mngs.dsp.demo_sig(fs=FS, t_sec=T_SEC, sig_type="tensorpac")
    # xx.shape: (8, 19, 20, 512)

    # Tensorpac
    (
        pha,
        amp,
        freqs_pha,
        freqs_amp,
        pac_tp,
    ) = mngs.dsp.utils.pac.calc_pac_with_tensorpac(xx, fs, t_sec=T_SEC)

    # GPU calculation with mngs.dsp.nn.ModulationIndex
    pha, amp = _reshape(pha), _reshape(amp)

    m = ModulationIndex(n_bins=18, fp16=True).to(device)

    pac_mngs = m(pha.to(device), amp.to(device))

    # pac_mngs = mngs.dsp.modulation_index(pha, amp).cpu().numpy()
    i_batch, i_ch = 0, 0
    pac_mngs = pac_mngs[i_batch, i_ch].squeeze().numpy()

    # Plots
    fig = mngs.dsp.utils.pac.plot_PAC_mngs_vs_tensorpac(
        pac_mngs, pac_tp, freqs_pha, freqs_amp
    )
    # fig = plot_PAC_mngs_vs_tensorpac(pac_mngs, pac_tp, freqs_pha, freqs_amp)
    mngs.io.save(fig, CONFIG["SDIR"] + "modulation_index.png")  # plt.show()

    # Close
    mngs.gen.close(CONFIG)

# EOF

"""
/home/ywatanabe/proj/entrance/mngs/nn/_ModulationIndex.py
"""



import matplotlib.pyplot as plt
import numpy as np


def map_ticks(ax, src, tgt, axis="x"):
    """
    Maps source tick positions or labels to new target labels on a matplotlib Axes object.
    Supports both numeric positions and string labels for source ticks ('src'), enabling the mapping
    to new target labels ('tgt'). This ensures only the specified target ticks are displayed on the
    final axis, enhancing the clarity and readability of plots.

    Parameters:
    - ax (matplotlib.axes.Axes): The Axes object to modify.
    - src (list of str or numeric): Source positions (if numeric) or labels (if str) to map from.
      When using string labels, ensure they match the current tick labels on the axis.
    - tgt (list of str): New target labels to apply to the axis. Must have the same length as 'src'.
    - axis (str): Specifies which axis to apply the tick modifications ('x' or 'y').

    Returns:
    - ax (matplotlib.axes.Axes): The modified Axes object with adjusted tick labels.

    Examples:
    --------
    Numeric Example:
        fig, ax = plt.subplots()
        x = np.linspace(0, 2 * np.pi, 100)
        y = np.sin(x)
        ax.plot(x, y)  # Plot a sine wave
        src = [0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi]  # Numeric src positions
        tgt = ['0', 'π/2', 'π', '3π/2', '2π']  # Corresponding target labels
        map_ticks(ax, src, tgt, axis="x")  # Map src to tgt on the x-axis
        plt.show()

    String Example:
        fig, ax = plt.subplots()
        categories = ['A', 'B', 'C', 'D', 'E']  # Initial categories
        values = [1, 3, 2, 5, 4]
        ax.bar(categories, values)  # Bar plot with string labels
        src = ['A', 'B', 'C', 'D', 'E']  # Source labels to map from
        tgt = ['Alpha', 'Beta', 'Gamma', 'Delta', 'Epsilon']  # New target labels
        map_ticks(ax, src, tgt, axis="x")  # Apply the mapping
        plt.show()
    """
    if len(src) != len(tgt):
        raise ValueError(
            "Source ('src') and target ('tgt') must have the same number of elements."
        )

    # Determine tick positions if src is string data
    if all(isinstance(item, str) for item in src):
        if axis == "x":
            all_labels = [label.get_text() for label in ax.get_xticklabels()]
        else:
            all_labels = [label.get_text() for label in ax.get_yticklabels()]

        # Find positions of src labels
        src_positions = [all_labels.index(s) for s in src if s in all_labels]
    else:
        # Use src as positions directly if numeric
        src_positions = src

    # Set the ticks and labels based on the specified axis
    if axis == "x":
        ax.set_xticks(src_positions)
        ax.set_xticklabels(tgt)
    elif axis == "y":
        ax.set_yticks(src_positions)
        ax.set_yticklabels(tgt)
    else:
        raise ValueError("Invalid axis argument. Use 'x' or 'y'.")

    return ax


def numeric_example():
    fig, axs = plt.subplots(2, 1, figsize=(10, 6))  # Two rows, one column

    # Original plot
    x = np.linspace(0, 2 * np.pi, 100)
    y = np.sin(x)
    axs[0].plot(x, y)  # Plot a sine wave on the first row
    axs[0].set_title("Original Numeric Labels")

    # Numeric src positions for ticks (e.g., multiples of pi) and target labels
    src = [0, np.pi / 2, np.pi, 3 * np.pi / 2, 2 * np.pi]
    tgt = ["0", "π/2", "π", "3π/2", "2π"]

    # Plot with mapped ticks
    axs[1].plot(x, y)  # Plot again on the second row for mapped labels
    map_ticks(axs[1], src, tgt, axis="x")
    axs[1].set_title("Mapped Numeric Labels")

    return fig


def string_example():
    fig, axs = plt.subplots(2, 1, figsize=(10, 6))  # Two rows, one column

    # Original plot with categorical string labels
    categories = ["A", "B", "C", "D", "E"]
    values = [1, 3, 2, 5, 4]
    axs[0].bar(categories, values)
    axs[0].set_title("Original String Labels")

    # src as the existing labels to change and target labels
    src = categories
    tgt = ["Alpha", "Beta", "Gamma", "Delta", "Epsilon"]

    # Plot with mapped string labels
    axs[1].bar(
        categories, values
    )  # Bar plot again on the second row for mapped labels
    map_ticks(axs[1], src, tgt, axis="x")
    axs[1].set_title("Mapped String Labels")

    return fig


# Execute examples
if __name__ == "__main__":
    fig_numeric = numeric_example()
    fig_string = string_example()

    plt.tight_layout()
    plt.show()



#!/usr/bin/env python3
# Time-stamp: "2024-02-17 12:38:40 (ywatanabe)"

from pprint import pprint
from time import sleep

# def get_params(model, tgt_name=None, sleep_sec=2, show=False):

#     name_shape_dict = {}
#     for name, param in model.named_parameters():
#         learnable = "Learnable" if param.requires_grad else "Freezed"

#         if (tgt_name is not None) & (name == tgt_name):
#             return param
#         if tgt_name is None:
#             # print(f"\n{param}\n{param.shape}\nname: {name}\n")
#             if show is True:
#                 print(
#                     f"\n{param}: {param.shape}\nname: {name}\nStatus: {learnable}\n"
#                 )
#                 sleep(sleep_sec)
#             name_shape_dict[name] = list(param.shape)

#     if tgt_name is None:
#         print()
#         pprint(name_shape_dict)
#         print()


def check_params(model, tgt_name=None, show=False):

    out_dict = {}

    for name, param in model.named_parameters():
        learnable = "Learnable" if param.requires_grad else "Freezed"

        if tgt_name is None:
            out_dict[name] = (param.shape, learnable)

        elif (tgt_name is not None) & (name == tgt_name):
            out_dict[name] = (param.shape, learnable)

        elif (tgt_name is not None) & (name != tgt_name):
            continue

    if show:
        for k, v in out_dict.items():
            print(f"\n{k}\n{v}")

    return out_dict



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-27 16:55:44 (ywatanabe)"

from ._AxiswiseDropout import AxiswiseDropout
from ._BNet import BNet, BNet_config
from ._ChannelGainChanger import ChannelGainChanger
from ._DropoutChannels import DropoutChannels
from ._Filters import (
    BandPassFilter,
    BandStopFilter,
    DifferentiableBandPassFilter,
    GaussianFilter,
    HighPassFilter,
    LowPassFilter,
)
from ._FreqGainChanger import FreqGainChanger
from ._Hilbert import Hilbert

# from ._FreqDropout import FreqDropout
from ._MNet_1000 import MNet_1000, MNet_config
from ._ModulationIndex import ModulationIndex
from ._PAC import PAC

# from ._PAC_dev import PAC_dev
from ._PSD import PSD
from ._ResNet1D import ResNet1D, ResNetBasicBlock
from ._SpatialAttention import SpatialAttention
from ._SwapChannels import SwapChannels
from ._TransposeLayer import TransposeLayer
from ._Wavelet import Wavelet



import torch
import warnings

def verify_n_gpus(n_gpus):
    if torch.cuda.device_count() < n_gpus:
        warnings.warn(
            f"N_GPUS ({n_gpus}) is larger "
            f"than n_gpus torch can acesses (= {torch.cuda.device_count()})"
            f"Please check $CUDA_VISIBLE_DEVICES and your setting in this script.",
            UserWarning,
        )
        return torch.cuda.device_count()

    else:
        return n_gpus



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-04 08:33:39 (ywatanabe)"

from ._distance import cdist, euclidean_distance
from ._geometric_median import geometric_median
from ._misc import cosine, nannorm, rebase_a_vec, three_line_lengths_to_coords



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-08 12:41:59 (ywatanabe)"#!/usr/bin/env python3


import warnings

import numpy as np
import pandas as pd
import torch
from mngs.general import torch_fn


def to_sktime_df(arr):
    """
    Convert a 3D numpy array into a DataFrame suitable for sktime.

    Parameters:
    arr (numpy.ndarray): A 3D numpy array with shape (n_samples, n_channels, seq_len)

    Returns:
    pandas.DataFrame: A DataFrame in sktime format
    """
    if len(arr.shape) != 3:
        raise ValueError("Input data must be a 3D array")

    n_samples, seq_len, n_channels = arr.shape

    # Initialize an empty DataFrame for sktime format
    sktime_df = pd.DataFrame(index=range(n_samples), columns=["dim_0"])

    # Iterate over each sample
    for i in range(n_samples):
        # Combine all channels into a single cell
        combined_series = pd.Series(
            {
                f"channel_{j}": pd.Series(arr[i, :, j])
                for j in range(n_channels)
            }
        )
        sktime_df.iloc[i, 0] = combined_series

    return sktime_df


@torch_fn
def to_segments(x, window_size, overlap_factor=1, dim=-1):
    stride = window_size // overlap_factor
    num_windows = (x.size(dim) - window_size) // stride + 1
    windows = x.unfold(dim, window_size, stride)
    return windows


if __name__ == "__main__":
    x, t, f = mngs.dsp.demo_sig()

    y = to_segments(x, 256)

    x = 100 * np.random.rand(16, 160, 1000)
    print(_normalize_time(x))

    x = torch.randn(16, 160, 1000)
    print(_normalize_time(x))

    x = torch.randn(16, 160, 1000).cuda()
    print(_normalize_time(x))



"""
This script does XYZ.
"""

# Imports

# Functions
def paste():
    import textwrap

    import pyperclip

    try:
        clipboard_content = pyperclip.paste()
        clipboard_content = textwrap.dedent(clipboard_content)
        exec(clipboard_content)
    except Exception as e:
        print(f"Could not execute clipboard content: {e}")


if __name__ == "__main__":
    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(sys, plt)

    import ipdb

    ipdb.set_trace()
    print("hello")  # copy this to your clipboar
    mngs.gen.paste()

    # Close
    mngs.gen.close(CONFIG)

# EOF

"""
/ssh:ywatanabe@444:/home/ywatanabe/proj/entrance/mngs/general/_paste.py
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-03-12 19:52:48 (ywatanabe)"

import re

import matplotlib
import matplotlib.pyplot as plt
import mngs
import numpy as np
import pandas as pd


def process_i_global(metrics_df):
    if metrics_df.index.name != "i_global":
        try:
            metrics_df = metrics_df.set_index("i_global")
        except KeyError:
            print(
                "Error: The DataFrame does not contain a column named 'i_global'. Please check the column names."
            )
        except Exception as e:
            print(f"An unexpected error occurred: {e}")
    else:
        print("The index is already set to 'i_global'.")
    metrics_df["i_global"] = metrics_df.index  # alias
    return metrics_df


def set_yaxis_for_acc(ax, key_plt):
    if re.search("[aA][cC][cC]", key_plt):  # acc, ylim, yticks
        ax.set_ylim(0, 1)
        ax.set_yticks([0, 0.5, 1.0])
    return ax


def plot_tra(ax, metrics_df, key_plt, lw=1, color="blue"):
    indi_step = mngs.gen.search(
        "^[Tt]rain(ing)?", metrics_df.step, as_bool=True
    )[0]
    step_df = metrics_df[indi_step]

    if len(step_df) != 0:
        ax.plot(
            step_df.index,  # i_global
            step_df[key_plt],
            label="Training",
            color=color,
            linewidth=lw,
        )
        ax.legend()

    return ax


def scatter_val(ax, metrics_df, key_plt, s=3, color="green"):
    indi_step = mngs.gen.search(
        "^[Vv]alid(ation)?", metrics_df.step, as_bool=True
    )[0]
    step_df = metrics_df[indi_step]
    if len(step_df) != 0:
        ax.scatter(
            step_df.index,
            step_df[key_plt],
            label="Validation",
            color=color,
            s=s,
            alpha=0.9,
        )
        ax.legend()
    return ax


def scatter_tes(ax, metrics_df, key_plt, s=3, color="red"):
    indi_step = mngs.gen.search("^[Tt]est", metrics_df.step, as_bool=True)[0]
    step_df = metrics_df[indi_step]
    if len(step_df) != 0:
        ax.scatter(
            step_df.index,
            step_df[key_plt],
            label="Test",
            color=color,
            s=s,
            alpha=0.9,
        )
        ax.legend()
    return ax


def vline_at_epochs(ax, metrics_df, color="grey"):
    # Determine the global iteration values where new epochs start
    epoch_starts = metrics_df[metrics_df["i_batch"] == 0].index.values
    epoch_labels = metrics_df[metrics_df["i_batch"] == 0].index.values
    ax.vlines(
        x=epoch_starts,
        ymin=-1e4,  # ax.get_ylim()[0],
        ymax=1e4,  # ax.get_ylim()[1],
        linestyle="--",
        color=color,
    )
    return ax


def select_ticks(metrics_df, max_n_ticks=4):
    # Calculate epoch starts and their corresponding labels for ticks
    unique_epochs = metrics_df["i_epoch"].drop_duplicates().values
    epoch_starts = (
        metrics_df[metrics_df["i_batch"] == 0]["i_global"]
        .drop_duplicates()
        .values
    )

    # Given the performance issue, let's just select a few epoch starts for labeling
    # We use MaxNLocator to pick ticks; however, it's used here to choose a reasonable number of epoch markers
    if len(epoch_starts) > max_n_ticks:
        selected_ticks = np.linspace(
            epoch_starts[0], epoch_starts[-1], max_n_ticks, dtype=int
        )
        # Ensure selected ticks are within the epoch starts for accurate labeling
        selected_labels = [
            metrics_df[metrics_df["i_global"] == tick]["i_epoch"].iloc[0]
            for tick in selected_ticks
        ]
    else:
        selected_ticks = epoch_starts
        selected_labels = unique_epochs
    return selected_ticks, selected_labels


def learning_curve(
    metrics_df,
    keys,
    title="Title",
    max_n_ticks=4,
    scattersize=3,
    linewidth=1,
    yscale="linear",
    spath=None,
):
    _plt, cc = mngs.plt.configure_mpl(plt, show=False)
    """
    Example:
        print(metrics_df)
        #                 step  i_global  i_epoch  i_batch      loss
        # 0       Training         0        0        0  0.717023
        # 1       Training         1        0        1  0.703844
        # 2       Training         2        0        2  0.696279
        # 3       Training         3        0        3  0.685384
        # 4       Training         4        0        4  0.670675
        # ...          ...       ...      ...      ...       ...
        # 123266      Test     66900      299      866  0.000067
        # 123267      Test     66900      299      867  0.000067
        # 123268      Test     66900      299      868  0.000067
        # 123269      Test     66900      299      869  0.000067
        # 123270      Test     66900      299      870  0.000068

        # [123271 rows x 5 columns]
    """
    metrics_df = process_i_global(metrics_df)
    selected_ticks, selected_labels = select_ticks(metrics_df)

    # fig, axes = plt.subplots(len(keys), 1, sharex=True, sharey=False)
    fig, axes = mngs.plt.subplots(len(keys), 1, sharex=True, sharey=False)
    axes = axes if len(keys) != 1 else [axes]

    axes[-1].set_xlabel("Iteration #")
    fig.text(0.5, 0.95, title, ha="center")

    for i_plt, key_plt in enumerate(keys):
        ax = axes[i_plt]
        ax.set_yscale(yscale)
        ax.set_ylabel(key_plt)

        ax = set_yaxis_for_acc(ax, key_plt)
        ax = plot_tra(ax, metrics_df, key_plt, lw=linewidth, color=cc["blue"])
        ax = scatter_val(
            ax, metrics_df, key_plt, s=scattersize, color=cc["green"]
        )
        ax = scatter_tes(
            ax, metrics_df, key_plt, s=scattersize, color=cc["red"]
        )

        # # Custom tick marks
        # ax = mngs.plt.ax.map_ticks(
        #     ax, selected_ticks, selected_labels, axis="x"
        # )

    if spath is not None:
        mngs.io.save(fig, spath)

    return fig


if __name__ == "__main__":

    plt, cc = mngs.plt.configure_mpl(plt)
    # lpath = "./scripts/ml/.old/pretrain_EEGPT_old/2024-01-29-12-04_eDflsnWv_v8/metrics.csv"
    lpath = "./scripts/ml/pretrain_EEGPT/[DEBUG] 2024-02-11-06-45_4uUpdfpb/metrics.csv"

    sdir, _, _ = mngs.gen.split_fpath(lpath)
    metrics_df = mngs.io.load(lpath)
    fig = learning_curve(
        metrics_df, title="Pretraining on db_v8", yscale="log"
    )
    # plt.show()
    mngs.io.save(fig, sdir + "learning_curve.png")



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-04 17:43:57 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/linalg/_geometric_median.py


"""
This script does XYZ.
"""

import torch
from geom_median.torch import compute_geometric_median
from mngs.gen import torch_fn

# @torch_fn
# def geometric_median(xx, dim=-1):
#     indi = [slice(None) for _ in range(xx.ndim)]
#     indi[dim] = slice(None)
#     xx[indi]  # how can I loop over the designated dim??

#     return compute_geometric_median(xx).median


@torch_fn
def geometric_median(xx, dim=-1):
    # Ensure dim is a positive index
    if dim < 0:
        dim = xx.ndim + dim

    # Create a list of slices to access all elements along each dimension
    indi = [slice(None)] * xx.ndim

    # Get the size of the dimension we want to loop over
    dim_size = xx.shape[dim]

    points = []
    # Loop over each index in the specified dimension
    for i in range(dim_size):
        indi[
            dim
        ] = i  # Set the slice for the current index in the target dimension
        slice_data = xx[tuple(indi)]  # Extract the data for the current index
        points.append(slice_data)

    return compute_geometric_median(points).median


if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!/usr/bin/env python
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import mngs


class MultiTaskLoss(nn.Module):
    """
    # https://openaccess.thecvf.com/content_cvpr_2018/papers/Kendall_Multi-Task_Learning_Using_CVPR_2018_paper.pdf

    Example:
        are_regression = [False, False]
        mtl = MultiTaskLoss(are_regression)
        losses = [torch.rand(1, requires_grad=True) for _ in range(len(are_regression))]
        loss = mtl(losses)
        print(loss)
        # [tensor([0.4215], grad_fn=<AddBackward0>), tensor([0.6190], grad_fn=<AddBackward0>)]
    """

    def __init__(self, are_regression=[False, False], reduction="none"):
        super().__init__()
        mngs.general.fix_seeds(np=np, torch=torch, show=False)
        n_tasks = len(are_regression)

        self.register_buffer("are_regression", torch.tensor(are_regression))

        # for the numercal stability, log(variables) are learned.
        self.log_vars = torch.nn.Parameter(torch.zeros(n_tasks))
        self.reduction = reduction

    def forward(self, losses):
        vars = torch.exp(self.log_vars).type_as(losses[0])
        stds = vars ** (1 / 2)
        coeffs = 1 / ((self.are_regression + 1) * vars)
        scaled_losses = [
            coeffs[i] * losses[i] + torch.log(stds[i]) for i in range(len(losses))
        ]
        return scaled_losses



#!/usr/bin/env python3


from collections import Counter

import numpy as np


def under_sample(y, replace=False):
    """
    Input:
        Labels
    Return:
        Indices

    Example:
        t = ['a', 'b', 'c', 'b', 'c', 'a', 'c']
        print(under_sample(t))
        # [5 0 1 3 4 6]
        print(under_sample(t))
        # [5 0 1 3 6 2]
    """

    # find the minority and majority classes
    class_counts = Counter(y)
    # majority_class = max(class_counts, key=class_counts.get)
    minority_class = min(class_counts, key=class_counts.get)

    # compute the number of sample to draw from the majority class using
    # a negative binomial distribution
    n_minority_class = class_counts[minority_class]
    n_majority_resampled = n_minority_class

    # draw randomly with or without replacement
    indices = np.hstack(
        [
            np.random.choice(
                np.flatnonzero(y == k),
                size=n_majority_resampled,
                replace=replace,
            )
            for k in class_counts.keys()
        ]
    )

    return indices


if __name__ == "__main__":
    t = np.array(["a", "b", "c", "b", "c", "a", "c"])
    print(under_sample(t))



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-08 00:06:40 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/plt/_subplots/AxisWrapper.py

from collections import OrderedDict
from contextlib import contextmanager
from functools import wraps

import mngs
import seaborn as sns
from mngs.gen import not_implemented

from ._to_sigma import to_sigma as _to_sigma


class AxisWrapper:
    """
    A wrapper class for a Matplotlib axis that collects plotting data.
    """

    # def __init__(self, axis, history, track):
    def __init__(self, axis, track):
        """
        Initialize the AxisWrapper with a given axis and history reference.
        """
        self.axis = axis
        # self._ax_history = history
        self._ax_history = OrderedDict()
        self.track = track
        self.id = 0

    def __getattr__(self, attr):
        """
        Wrap the axis attribute access to collect plot calls or return the attribute directly.
        """
        original_attr = getattr(self.axis, attr)

        if callable(original_attr):

            @wraps(original_attr)
            def wrapper(
                *args, track=None, id=None, n_xticks=4, n_yticks=4, **kwargs
            ):
                results = original_attr(*args, **kwargs)

                # if attr in ["plot", "scatter", "plot_with_ci"]:
                #     self.axis = mngs.plt.ax.set_n_ticks(
                #         self.axis, n_xticks=n_xticks, n_yticks=n_yticks
                #     )

                self._track(track, id, attr, args, kwargs)

                return results

            return wrapper

        else:
            return original_attr

    ################################################################################
    ## Tracking
    ################################################################################
    def _track(self, track, id, method_name, args, kwargs):
        if track is None:
            track = self.track
        if track:
            id = id if id is not None else self.id
            self.id += 1
            self._ax_history[id] = (id, method_name, args, kwargs)

    @contextmanager
    def _no_tracking(self):
        """Context manager to temporarily disable tracking."""
        original_track = self.track
        self.track = False
        try:
            yield
        finally:
            self.track = original_track

    @property
    def history(self):
        return {k: self._ax_history[k] for k in self._ax_history}

    def reset_history(self):
        self._ax_history = {}

    def to_sigma(self):
        """
        Export tracked plotting data to a DataFrame in SigmaPlot format.
        """
        df = _to_sigma(self.history)
        self.reset_history()

        return df

    ################################################################################
    ## Plotting methods (other than matplotlib's original methods)
    ################################################################################
    def imshow2d(
        self,
        arr_2d,
        cbar=True,
        cbar_label=None,
        cbar_shrink=1.0,
        cbar_fraction=0.046,
        cbar_pad=0.04,
        cmap="viridis",
        aspect="auto",
        vmin=None,
        vmax=None,
        track=None,
        id=None,
        **kwargs,
    ):
        # Method name
        method_name = "imshow2d"

        # Plotting
        with self._no_tracking():
            self.axis = mngs.plt.ax.imshow2d(
                self.axis,
                arr_2d,
                cbar=cbar,
                cbar_label=cbar_label,
                cbar_shrink=cbar_shrink,
                cbar_fraction=cbar_fraction,
                cbar_pad=cbar_pad,
                cmap=cmap,
                aspect=aspect,
                vmin=vmin,
                vmax=vmax,
            )

        # Tracking
        out = arr_2d
        self._track(track, id, method_name, out, None)

    def rectangle(self, xx, yy, ww, hh, track=None, id=None, **kwargs):
        # Method name
        method_name = "rectangle"

        # Plotting
        # with self._no_tracking():
        self.axis = mngs.plt.ax.rectangle(self.axis, xx, yy, ww, hh, **kwargs)

        # Tracking
        out = None
        self._track(track, id, method_name, out, None)

    def plot_with_ci(
        self, xx, mean, std, label=None, alpha=0.5, track=None, id=None
    ):
        # Method name
        method_name = "plot_with_ci"

        # Plotting
        with self._no_tracking():
            self.axis = mngs.plt.ax.plot_with_ci(
                self.axis, xx, mean, std, label=label, alpha=alpha
            )

        # Tracking
        out = (xx, mean, std)
        self._track(track, id, method_name, out, None)

    def raster(self, positions, time=None, track=None, id=None, **kwargs):
        # Method name
        method_name = "raster"

        # Plotting
        with self._no_tracking():
            self.axis, df = mngs.plt.ax.raster(
                self.axis, positions, time=time, **kwargs
            )
        out = df

        # Tracking
        self._track(track, id, method_name, out, None)

    @not_implemented
    def joyplot(
        self,
        track=None,
        id=None,
    ):
        # Method name
        method_name = "joyplot"

        # Plotting
        __import__("ipdb").set_trace()

        # Tracking
        out = None
        self._track(track, id, method_name, out, None)

    ################################################################################
    ## Seaborn-wrappers
    ################################################################################
    def _sns_base(self, method_name, *args, track=None, id=None, **kwargs):
        actual_method_name = method_name.split("sns_")[-1]

        with self._no_tracking():
            plot_func = getattr(sns, actual_method_name)
            self.axis = plot_func(*args, **kwargs)

        # Track the plot if required
        self._track(track, id, method_name, args, kwargs)

    def sns_barplot(self, *args, track=None, id=None, **kwargs):
        self._sns_base("sns_barplot", *args, track=track, id=id, **kwargs)

    def sns_boxplot(self, *args, track=None, id=None, **kwargs):
        self._sns_base("sns_boxplot", *args, track=track, id=id, **kwargs)

    def sns_heatmap(self, *args, track=None, id=None, **kwargs):
        self._sns_base("sns_heatmap", *args, track=track, id=id, **kwargs)

    def sns_histplot(self, *args, track=None, id=None, **kwargs):
        self._sns_base("sns_histplot", *args, track=track, id=id, **kwargs)

    def sns_kdeplot(self, *args, track=None, id=None, **kwargs):
        self._sns_base("sns_kdeplot", *args, track=track, id=id, **kwargs)

    def sns_lineplot(self, *args, track=None, id=None, **kwargs):
        self._sns_base("sns_lineplot", *args, track=track, id=id, **kwargs)

    def sns_pairplot(self, *args, track=None, id=None, **kwargs):
        self._sns_base("sns_pairplot", *args, track=track, id=id, **kwargs)

    def sns_scatterplot(self, *args, track=None, id=None, **kwargs):
        self._sns_base("sns_scatterplot", *args, track=track, id=id, **kwargs)

    def sns_violinplot(self, *args, track=None, id=None, **kwargs):
        self._sns_base("sns_violinplot", *args, track=track, id=id, **kwargs)

    def sns_jointplot(self, *args, track=None, id=None, **kwargs):
        self._sns_base("sns_jointplot", *args, track=track, id=id, **kwargs)

    ################################################################################
    ## Adjusting methods
    ################################################################################
    def set_xyt(
        self,
        xlabel=None,
        ylabel=None,
        title=None,
    ):
        self.axis = mngs.plt.ax.set_xyt(
            self.axis,
            xlabel=xlabel,
            ylabel=ylabel,
            title=title,
        )

    def set_supxyt(
        self,
        xlabel=None,
        ylabel=None,
        title=None,
    ):
        self.axis = mngs.plt.ax.set_supxyt(
            self.axis,
            xlabel=xlabel,
            ylabel=ylabel,
            title=title,
        )

    def set_ticks(
        self,
        xvals=None,
        xticks=None,
        yvals=None,
        yticks=None,
        **kwargs,
    ):

        self.axis = mngs.plt.ax.set_ticks(
            self.axis,
            xvals=xvals,
            xticks=xticks,
            yvals=yvals,
            yticks=yticks,
        )

    def set_n_ticks(self, n_xticks=4, n_yticks=4):
        self.axis = mngs.plt.ax.set_n_ticks(
            self.axis, n_xticks=n_xticks, n_yticks=n_yticks
        )

    def hide_spines(
        self,
        top=True,
        bottom=True,
        left=True,
        right=True,
        ticks=True,
        labels=True,
    ):
        self.axis = mngs.plt.ax.hide_spines(
            self.axis,
            top=top,
            bottom=bottom,
            left=left,
            right=right,
            ticks=ticks,
            labels=labels,
        )

    def extend(self, x_ratio=1.0, y_ratio=1.0):
        self.axis = mngs.plt.ax.extend(
            self.axis, x_ratio=x_ratio, y_ratio=y_ratio
        )

    def shift(self, dx=0, dy=0):
        self.axis = mngs.plt.ax.shift(self.axis, dx=dx, dy=dy)



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-26 20:56:18 (ywatanabe)"


"""
This script does XYZ.
"""


"""
Imports
"""
import os
import sys

import matplotlib.pyplot as plt
import mngs
import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F

"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""
import numpy as np


def crop(sig_2d, window_length, overlap_factor=0.0, axis=-1, time=None):
    """
    Crops the input signal into overlapping windows of a specified length,
    allowing for an arbitrary axis and considering a time vector.

    Parameters:
    - sig_2d (numpy.ndarray): The input sig_2d array to be cropped. Can be multi-dimensional.
    - window_length (int): The length of each window to crop the sig_2d into.
    - overlap_factor (float): The fraction of the window that consecutive windows overlap. For example, an overlap_factor of 0.5 means 50% overlap.
    - axis (int): The time axis along which to crop the sig_2d.
        - time (numpy.ndarray): The time vector associated with the signal. Its length should match the signal's length along the cropping axis.

    Returns:
    - cropped_windows (numpy.ndarray): The cropped signal windows. The shape depends on the input shape and the specified axis.
    """
    # Ensure axis is in a valid range
    if axis < 0:
        axis += sig_2d.ndim
    if axis >= sig_2d.ndim or axis < 0:
        raise ValueError(
            "Invalid axis. Axis out of range for sig_2d dimensions."
        )

    if time is not None:
        # Validate the length of the time vector against the signal's dimension
        if sig_2d.shape[axis] != len(time):
            raise ValueError(
                "Length of time vector does not match signal's dimension along the specified axis."
            )

    # Move the target axis to the last position
    axes = np.arange(sig_2d.ndim)
    axes[axis], axes[-1] = axes[-1], axes[axis]
    sig_2d_permuted = np.transpose(sig_2d, axes)

    # Compute the number of windows and the step size
    seq_len = sig_2d_permuted.shape[-1]
    step = int(window_length * (1 - overlap_factor))
    n_windows = max(
        1, ((seq_len - window_length) // step + 1)
    )  # Ensure at least 1 window

    # Crop the sig_2d into windows
    cropped_windows = []
    cropped_times = []
    for i in range(n_windows):
        start = i * step
        end = start + window_length
        cropped_windows.append(sig_2d_permuted[..., start:end])
        if time is not None:
            cropped_times.append(time[start:end])

    # Convert list of windows back to numpy array
    cropped_windows = np.array(cropped_windows)
    cropped_times = np.array(cropped_times)

    # Move the last axis back to its original position if necessary
    if axis != sig_2d.ndim - 1:
        # Compute the inverse permutation
        inv_axes = np.argsort(axes)
        cropped_windows = np.transpose(cropped_windows, axes=inv_axes)

    if time is None:
        return cropped_windows
    else:
        return cropped_windows, cropped_times


def main():
    import random

    FS = 128
    N_CHS = 19
    RECORD_S = 13
    WINDOW_S = 2
    FACTOR = 0.5

    # To pts
    record_pts = int(RECORD_S * FS)
    window_pts = int(WINDOW_S * FS)

    # Demo signal
    sig2d = np.random.rand(N_CHS, record_pts)
    time = np.arange(record_pts) / FS

    # Main
    xx, tt = crop(sig2d, window_pts, overlap_factor=FACTOR, time=time)

    print(f"sig2d.shape: {sig2d.shape}")
    print(f"xx.shape: {xx.shape}")

    # Validation
    i_seg = random.randint(0, len(xx) - 1)
    start = int(i_seg * window_pts * FACTOR)
    end = start + window_pts
    assert np.allclose(sig2d[:, start:end], xx[i_seg])


if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF
